import numpy as np
from time import time


def mp_corr(iproc, nf, chn, srch, rcr, lind, nq, quc, quce):
    def verbose0(reg):
        return
        # some output for testing purposes
        print('\nn: ', n)
        print('reg: ', reg)
        print('sst {}\nsr:    {}\nsl:    {}'.format(sst, sr, sl))
        print('nk: ', nk)
        print('datreg: ', datreg[reg])
        print('corf: ', corf)
    def verbose1(reg):
        return
        # some output for testing purposes
        print('\nn: ', n)
        print('reg: ', reg)
        print('sst {}\nsr:    {}\nsl:    {}'.format(sst, sr, sl))
        print('srr {}\nsll {}'.format(srr[reg], sll[reg]))
        print('nk: ', nk)
        print('corf: ', corf)
        print('datreg: ', datreg[reg])
    
    def running_mean( m, s, x, k):
        mt = 1.*m
        m += (x-m)/k
        s += (x-mt)*(x-m)

    def correlator(reg, matrx):

        if n < chn:
            if n:
                nk[:n] += 1
            for cindex in xnq:
                sr[cindex,:n] = np.roll(sr[cindex,:n],1)
                sr[cindex,0] = sr[cindex,1] + sst[reg][cindex]
                sst[reg][cindex] = matrx[cindex].mean()
                sl[cindex,:n] += sst[reg][cindex]

            for cindex in xnq:
                x = np.dot(matrx[cindex], datreg[reg][cindex][:,:n]) / lind[cindex]
                running_mean(corf[cindex,:n], dcorf[cindex,:n], x, nk[:n])
                datreg[reg][cindex] = np.concatenate(
                    (matrx[cindex][:,None], datreg[reg][cindex][:,:chn-1]), axis=1)
            verbose0(reg)
            if n%2:
                for cindex in xnq:
                    matrx[cindex] = (datreg[reg][cindex][:,0] + datreg[reg][cindex][:,1])/2.

                correlator2(1,matrx)
        else:
            nk[:chn] += 1
            for cindex in xnq:
                sr[cindex,:chn] = np.roll(sr[cindex,:chn],1)
                sr[cindex,0] = sr[cindex,1] + sst[reg][cindex]
                sst[reg][cindex] = matrx[cindex].mean()
                sl[cindex,:chn] += sst[reg][cindex]

                x = np.dot(matrx[cindex], datreg[reg][cindex]) / lind[cindex]
                running_mean(corf[cindex,:chn], dcorf[cindex,:chn], x, nk[:chn])

                datreg[reg][cindex] = np.concatenate(
                    (matrx[cindex][:,None], datreg[reg][cindex][:,:chn-1]), axis=1)

            verbose0(reg)
            if n%2:
                for cindex in xnq:
                    matrx[cindex] = (datreg[reg][cindex][:,0] + datreg[reg][cindex][:,1])/2.

                correlator2(1,matrx)

    def correlator2(reg, matrx):
        condition = ((reg+1)<srch and (n+1)%2**(reg+1)==0) # if condition is true, move to
        kreg = int((n+1)/2**reg)                           # the next register
       # print('\n\ncondition: ', condition)
       # print('kreg: ', kreg)

        if 1 < kreg <= chn:
            for cindex in xnq:
                srr[reg][cindex,:kreg-1] = np.roll(srr[reg][cindex,:kreg-1],1)
                srr[reg][cindex,0] = srr[reg][cindex,1] + sst[reg][cindex]
                sst[reg][cindex] = matrx[cindex].mean()
                sll[reg][cindex,:kreg-1] += sst[reg][cindex]

        elif kreg > chn:
            for cindex in xnq:
                srr[reg][cindex,:chn] = np.roll(srr[reg][cindex,:chn],1)
                srr[reg][cindex,0] = srr[reg][cindex,1] + sst[reg][cindex]
                sst[reg][cindex] = matrx[cindex].mean()
                sll[reg][cindex,:chn] += sst[reg][cindex]

        if  (chn2+1) < kreg <= chn:
            inb = chn2*(reg+1)
            ine = chn2*reg + kreg - 1
            sl[:,inb:ine] = sll[reg][:,chn2:kreg-1]
            sr[:,inb:ine] = srr[reg][:,chn2:kreg-1]

            nk[inb:ine] += 1
            for cindex in xnq:
                x = np.dot(matrx[cindex],datreg[reg][cindex][:,chn2:kreg-1]) / lind[cindex]
                #print('x: ', x)
                running_mean(corf[cindex,inb:ine], dcorf[cindex,inb:ine], x, nk[inb:ine])
                datreg[reg][cindex] = np.concatenate(
                    (matrx[cindex][:,None], datreg[reg][cindex][:,:chn-1]), axis=1)
            verbose1(reg)

            if condition:
                for cindex in xnq:
                    matrx[cindex] = (datreg[reg][cindex][:,0] + datreg[reg][cindex][:,1])/2.
                reg += 1
                correlator2(reg, matrx)

        elif kreg > chn:
            inb = chn2*(reg+1)
            ine = chn2*(reg+2)
            sl[:,inb:ine] = sll[reg][:,chn2:chn]
            sr[:,inb:ine] = srr[reg][:,chn2:chn]

            nk[inb:ine] += 1
            for cindex in xnq:
                x = np.dot(matrx[cindex], datreg[reg][cindex][:,chn2:chn]) / lind[cindex]
                #print('x: ', x)
                running_mean(corf[cindex,inb:ine], dcorf[cindex,inb:ine], x, nk[inb:ine])
                datreg[reg][cindex] = np.concatenate(
                    (matrx[cindex][:,None], datreg[reg][cindex][:,:chn-1]), axis=1)

            verbose1(reg)
            if condition:
                for cindex in xnq:
                    matrx[cindex] = (datreg[reg][cindex][:,0] + datreg[reg][cindex][:,1])/2.
                reg += 1
                correlator2(reg, matrx)

        else:
            for cindex in xnq:
                sst[reg][cindex] = matrx[cindex].mean()
                datreg[reg][cindex] = np.concatenate(
                    (matrx[cindex][:,None], datreg[reg][cindex][:,:chn-1]), axis=1)

            verbose1(reg)
            if condition:
                 for cindex in xnq:
                      matrx[cindex] = (datreg[reg][cindex][:,0] + datreg[reg][cindex][:,1])/2.
                 reg += 1
                 correlator2(reg, matrx)


    # finished initializing part
    tcalc = time()
    chn2 = int(chn/2)
    datregt = []
    datreg = []
    xnq = range(nq)

    for ir in range(srch):
        for iq in xnq:
            datregt.append(np.zeros((lind[iq],chn), dtype=np.float32))
        datreg.append(datregt)
        datregt = []
    del datregt

    corf = np.zeros((nq,rcr), dtype=np.float32)
    dcorf = corf.copy()
    sl = corf.copy()
    sr = corf.copy()
    nk = np.zeros(rcr, dtype=np.float32)

    sll = []
    srr = []
    sst = []
    for ir in range(srch):
       sll.append(np.zeros((nq,chn), dtype=np.float32))
       srr.append(np.zeros((nq,chn), dtype=np.float32))
       sst.append(np.zeros(nq))

    #END of declaring and initializing variables####
    n = 0
    while n < nf:
        chunk = quc.get()
        chunk_size = chunk[0].shape[0]
        ni = 0
        while ni < chunk_size:
            matr = [qchunk[ni] for qchunk in chunk]
            correlator(0,matr)
            ni += 1
            n += 1


    # print('corf: ', corf)
    # print('\nsr: ', sr)
    # print('\nsl: ', sl)
    # print('\nnk: ',nk)
    
    #END OF MAIN LOOP put results to output queue
    quc.close()
    quc.join_thread()
    tcalc = time() - tcalc
    quce.put([corf.T, dcorf.T, nk, sr.T, sl.T, tcalc])



import pyFAI

class Eiger500k(pyFAI.detectors.Eiger):

    def __init__(self, *args, **kwargs):

        super().__init__(*args, **kwargs)
        self.aliases = ['Eiger500k']
        self.dim = self.shape = (514, 1030)



from matplotlib import pyplot as plt
import numpy as np
from Xana.Xplot.niceplot import niceplot

import matplotlib
import pickle
import pandas as pd
import lmfit
from Xana.Xfit.fit_basic import fit_basic
import ipywidgets as widgets
from ipywidgets import Layout, Button, Box, Text
from IPython.display import display, clear_output
import qgrid
from Xana.misc.logind import gen_logind

class ColeCole:

    def __init__(self, d, row, dofit=False, omega_max=np.inf, gpmax=np.inf):
        self.d = d
        self.dofit = dofit
        self.omega_max = omega_max
        self.gpmax = gpmax
        self.fitg0 = False
        self.params = row
        self.pl_data = None
        self.pl_fit = None
        self.ax = None
        self.xl = None
        self.normfactor = 1.

    @property
    def params(self):
        return self.__params

    @params.setter
    def params(self, vals):
        if isinstance(vals, lmfit.parameter.Parameters):
            p = vals
        elif vals is None or self.dofit:
            p = lmfit.Parameters()
            p.add('gosc', value=0, min=0)
            p.add('gmax', value=2, vary=0)
        elif isinstance(vals, (pd.DataFrame, pd.Series)):
            p = lmfit.Parameters()
            p.add('gosc', value=vals['gosc'], vary=False)
            p.add('gmax', value=vals['gmax'],vary=False)
            p.add('g0', value=vals['g0'],vary=False)
            p.add('gslp', value=vals['gslp'],vary=False)
            tmp = vals[['g0','gslp']].values
            if not np.isnan(tmp[0]):
                self.fitg0 = True
        self.__params = p

    @staticmethod
    def _fit_func(x, gosc):
        return np.sqrt(np.abs(x ** 2 - gosc * x))

    def fit(self, fitg0=False, g0range=[0,200]):

        def residuals(pars, x, data=None ):

            v = pars.valuesdict()
            fit = ColeCole._fit_func(x, v['gosc'])
            resid = (data - fit)
            # resid /= data
            return resid

        d = self.d
        omega_max = self.omega_max

        x = 'omega in rad/s'
        y = ["G' in Pa", "G'' in Pa"]

        gp = d[y[0]].values
        gpp = d[y[1]].values
        maxind = np.argmax(gp[d[x] < self.gpmax])

        gp_max = gp[maxind]
        self.params['gosc'].set(2*gp_max)
        self.params['gmax'].set(gp_max, vary=0)
        ind = d[x] < omega_max

        out = lmfit.minimize(residuals, self.params, args=(gp[ind],),
                             kws={'data': gpp[ind],}, nan_policy='omit')
        # print(lmfit.fit_report(out))
        self.params = out.params
        self.dofit = True

        if fitg0:
            self.params.add('g0', value=0, vary=0)
            self.params.add('gslp', value=0, vary=0)

            gosc = self.params['gosc'].value
            omega = d[x]
            ind2 = (omega>g0range[0]) & (omega<g0range[1])
            x = gp[ind2]
            y = gpp[ind2]
            # dy = residuals(self.params, x, y)
            p = fit_basic(x, y, model='lin', init={'m':(-gosc/2, None, 0),
                                                   'b':(gosc*2, 0, None)})[2]
            p = p.params
            for k in p:
                err = p[k].stderr
                p[k].stderr = err if err else 0
            self.params['g0'].value = -1. * p['b'].value / p['m'].value
            self.params['g0'].stderr = np.sqrt((p['b'].stderr/p['m'].value)**2
                                               + (p['b'].value*p['m'].stderr/p['m'].value**2)**2)
            self.params['gslp'].value = p['m'].value
            self.params['gslp'].stderr = p['b'].value
            self.fitg0 = True

    def plot(self, ax, xl=[0.01,500], normto='gosc', **kwargs):

        d = self.d
        self.ax = ax
        self.xl = xl
        # omega_max = self.omega_max

        x = 'omega in rad/s'
        y = ["G' in Pa", "G'' in Pa"]

        ind = np.where((d[x] < 90) & (d[x]>xl[0]) & (d[x]<xl[1]))[0]
        self.xl = [d[x][ind[0]], d[x][ind[-1]]]

        gp = d[y[0]].values[ind]
        gpp = d[y[1]].values[ind]
        norm = 1.
        if isinstance(normto, str):
            if normto in self.params:
                norm = self.params[normto].value
                self.normfactor = norm

        self.pl_dat, = ax.plot(gp / norm, gpp / norm,  'o', markersize=3)

        ax.set_ylabel("G'' in Pa")
        ax.set_xlabel("G' in Pa")
        ax.set_ylim([0, 1.5])
        ax.set_xlim([0, 3])

        if self.dofit:
            self.plot_fit(fitg0=self.fitg0)

    def plot_fit(self, xl=None, fitg0=True):
        ax = self.ax
        norm = self.normfactor
        # textstr = "$G_{{osc}}$ = {:.2g} +/- {:.2g} Pa".format(
        #     self.params['gosc'].value, self.params['gosc'].stderr)
        xf = np.linspace(0, self.params['gosc'].value, 200)
        yf = ColeCole._fit_func(xf, self.params['gosc'])
        self.pl_fit, = ax.plot(xf / norm, yf / norm, label=None,
                               color=self.pl_dat.get_color())
        if fitg0:
            xf = np.linspace(norm, 3*norm, 200)
            yf = self.params['gslp'].value * (xf  - self.params['g0'].value)
            # textstr = "$G_{{0}}$ = {:.2g} +/- {:.2g} Pa".format(
            #     self.params['g0'].value, self.params['g0'].stderr)
            ax.plot(xf / norm, yf / norm, label=None,
                    color=self.pl_dat.get_color())
            # ax.legend()

class Maxwell:

    def __init__(self, d, row, dofit=False, omega_max=np.inf):
        self.d = d
        self.dofit = dofit
        self.omega_max = omega_max
        self.params = row
        self.pl_dat = None
        self.pl_fit = None
        self.ax = None
        self.xl = None
        self.plot_ind = [1, 1]

    @property
    def params(self):
        return self.__params

    @params.setter
    def params(self, vals):
        if isinstance(vals, lmfit.parameter.Parameters):
            p = vals
        elif vals is None or self.dofit:
            p = lmfit.Parameters()
            p.add('eta', value=1, min=0)
            p.add('lmbd', value=10, min=0)
            p.add('gp', value=0.1, expr='eta/lmbd')
        elif isinstance(vals, (pd.DataFrame, pd.Series)):
            p = lmfit.Parameters()
            p.add('eta', value=vals['eta'], vary=False)
            p.add('lmbd', value=vals['lmbd'],vary=False)
            p.add('gp', value=0.1, expr='eta/lmbd')
        self.__params = p

    @staticmethod
    def _mw_gp(x, eta, lbd):
        return (eta * lbd * x ** 2) / (1 + x ** 2 * lbd ** 2)

    @staticmethod
    def _mw_gpp(x, eta, lbd):
        return (eta * x) / (1 + x ** 2 * lbd ** 2)

    def fit(self):

        def residuals(pars, x, gp, gpp):
            """2D Residual function to minimize
            """
            v = pars.valuesdict()
            resid = np.zeros((2, gp.size))
            resid[0] = (gp - Maxwell._mw_gp(x, v['eta'], v['lmbd']))
            resid[1] = (gpp - Maxwell._mw_gpp(x, v['eta'], v['lmbd']))
            return np.squeeze(resid.flatten())

        d = self.d
        xn = 'omega in rad/s'
        yn = ["G' in Pa", "G'' in Pa"]
        omega = d[xn]
        gp = d[yn[0]]
        gpp = d[yn[1]]
        indf = omega < self.omega_max
        omega = omega[indf]
        gp = gp[indf]
        gpp = gpp[indf]

        out = lmfit.minimize(residuals, self.params,
                             args=(omega, gp, gpp),
                             nan_policy='omit')

        self.params = out.params
        self.dofit = True

    def plot(self, ax, omega_max=np.inf, xl=[0.01,500], npoints=100, logspace=False,
             markersize=3, linestyle='', plot_gp=True, plot_gpp=True, **kwargs):

        self.ax = ax
        self.xl = xl
        d = self.d
        fill_style = {0: 'none', 1: 'full'}

        xn = 'omega in rad/s'
        yn = ["G' in Pa", "G'' in Pa"]
        self.plot_ind = np.array([plot_gp, plot_gpp], dtype='int8')
        yn = [yn[i] for i,xi in enumerate(self.plot_ind) if xi]

        fc = 0
        color = None
        x = d[xn].values
        ind = np.where((x>xl[0]) & (x<xl[1]) & (x<omega_max))[0]
        if logspace:
            ind = ind[gen_logind(len(ind), min([npoints,len(ind)]))]
        else:
            ind = ind[np.unique(np.linspace(0,len(ind)-1,npoints).astype(np.int))]
        self.xl = [x[ind[0]], x[ind[-1]]]

        for yni in yn:
            y = d[yni].values
            pl, = ax.plot(x[ind], y[ind], 'o', linestyle=linestyle, fillstyle=fill_style[fc], color=color, markersize=markersize)
            if fc == 0:
                color = pl.get_color()
                fc = 1

        ax.set_xlabel('$\\omega\\, (\\mathrm{{rad/s}})$')
        ax.set_ylabel("$G',\\, G''\\, (\\mathrm{{Pa}})$")

        ax.set_xscale('log')
        ax.set_yscale('log')

        if self.dofit:
            self.plot_fit()

    def plot_fit(self, xl=None):
        if xl is None:
            xl = self.xl
        xf = np.logspace(np.log10(xl[0]), np.log10(xl[1]), 200)
        p = self.params
        gpf = Maxwell._mw_gp(xf, p['eta'].value, p['lmbd'].value)
        gppf = Maxwell._mw_gpp(xf, p['eta'].value, p['lmbd'].value)
        if self.plot_ind[0]:
            self.ax.plot(xf, gpf, '-',
                         color=self.ax.get_lines()[-1].get_color())
        if self.plot_ind[1]:
            self.ax.plot(xf, gppf, '-',
                         color=self.ax.get_lines()[-1].get_color())


class Rheo:
    """Base class to analyze rheology data.
    """

    def __init__(self, df, datdir='./data'):

        self.df = df
        self.datdir = datdir
        self.register = {
            'all': None,
            'dynamic_moduli': 1,
            'flow_curve': 2,
        }
        self.plots = {
            'dynamic moduli': 1,
            'dynamic viscosity': 0,
            'cole-cole': 1,
            'flow curve eta': 1,
            'flow curve tau': 0,
            'temperature': 0,
        }
        self.nplots = sum([x for x in self.plots.values()])
        self.ax = None
        self.fig = None
        self.params = None
        self.table = None
        self.interactive = False
        self.data = {}

    @staticmethod
    def read_rheo_data(filename):
        return pd.read_csv(filename, sep=';', na_values=' ')

    def dname(self, row, counter=None, ):
        fname = row['file'].replace('.rwd', '').replace('.csv','')
        if counter:
            fname = row['datdir'] +'/'+ fname + '_{:02d}.dat'.format(counter)
        else:
            fname = row['datdir'] +'/'+ fname + '.dat'
        return fname

    @staticmethod
    def plot_dynamic_viscosity(d, ax, **kwargs):
        x = 'omega in rad/s'
        y = '|Eta*| in Pas'
        xvals = d[x].values
        yvals = d[y].values
        omega_max = kwargs.get('omega_max', np.inf)
        ind = np.where(xvals < omega_max)
        pl, = ax.plot(xvals[ind], yvals[ind], marker='o', )
        ax.set_xscale('log')
        ax.set_xlabel(x)
        ax.set_ylabel(y)
        return pl

    @staticmethod
    def plot_flow_curve_tau(d, ax, **kwargs):
        x = 'GP in 1/s'
        y = 'Tau in Pa'
        ax.plot(d[x].values, d[y].values, marker='o', )
        ax.set_xlabel(x)
        ax.set_ylabel(y)

    @staticmethod
    def plot_flow_curve_eta(d, ax, rngx=(0.02, 0.2), rngy=(0,100), **kwargs):
        xn = 'GP in 1/s'
        yn = 'Eta in Pas'
        x = d[xn].values
        y = d[yn].values
        ind = x > 0.01
        x = x[ind]
        y = y[ind]
        ind = (x > rngx[0]) & (x < rngx[1]) & (y > rngy[0]) & (y < rngy[1])
        eta = np.mean(y[ind])
        deta = np.std(y[ind])

        ax.plot(x, y, marker='o', label=r'$\eta_0$ = {:.2f} +/- {:.2f} Pas'.format(eta, deta))
        ax.set_xlabel(xn)
        ax.set_ylabel(yn)
        ax.set_xscale('log')
        ax.set_yscale('log')
        ax.vlines(rngx[0], eta - deta * 5, eta + deta * 5)
        ax.vlines(rngx[1], eta - deta * 5, eta + deta * 5)
        ax.legend()
        return eta, deta

    @staticmethod
    def plot_viscosity_aging(d, ax, **kwargs):
        x = 't_seg in s'
        y = '|Eta*| in Pas'
        ax.plot(d[x].values, d[y].values, marker='o', )
        ax.set_xlabel(x)
        ax.set_ylabel(y)

    @staticmethod
    def plot_temperature(d, ax, marker='o', **kwargs):
        x = 't in s'
        y = 'T in °C'
        ax.plot(d[x].values, d[y].values, marker=marker, **kwargs)
        ax.set_xlabel(x)
        ax.set_ylabel(y)

    def init_figure(self, ):

        nplots = sum([int(x) for x in self.plots.values()])
        if nplots != self.nplots:
            self.nplots = nplots
            if self.fig is not None:
                self.fig.clf()
            self.ax = None

        if self.ax is None:
            ncols = 1 if self.nplots <= 1 else 2
            nrows = (self.nplots + 1) // 2
            self.fig, ax = plt.subplots(nrows, ncols, figsize=(ncols * 4, nrows * 3), constrained_layout=True)
            with self.output:
                clear_output(wait=True)
                display(self.fig)
            self.ax = ax.flatten() if isinstance(ax, np.ndarray) else [ax, ]
        else:
            for axi in self.ax:
                axi.clear()


    def add_columns(self, columns):
        for c in columns[::-1]:
            if not c in self.df.columns:
                self.df.insert(7, c, np.nan)
            if self.table is not None:
                if not c in self.table.df.columns:
                    self.table.df.insert(7, c, np.nan)

    @staticmethod
    def set_rowparams(row, params):
        for p in params:
            row[f'{p}'] = params[p].value
            row[f'd_{p}'] = params[p].stderr

    def plot(self, df=None, dofit=True, doplot=True, cole_omega_max=30,
             maxwell_omega_max=40, g0range=[0,200], fitg0=False, defaultQ=1,
             gpmax=np.inf, plot_fit=False, plot_kws={}):

        if df is None:
            df = self.df

        ax = self.ax
        h_plot = []
        self.data = {pn:{} for pn in self.plots.keys()}
        for index, row in df[df['plot']].iterrows():

            if '18' in row['datdir']:
                register = {
                    'all': None,
                    'dynamic_moduli': 2,
                    'flow_curve': 3,
                }
            elif 'data' in row['datdir']:
                register = {
                    'all': None,
                    'dynamic_moduli': 1,
                    'flow_curve': 2,
                }
            elif 'p6380' in row['datdir']:
                register = {
                    'all': None,
                    'dynamic_moduli': 1,
                    'flow_curve': 2,
                }
            else:
                continue

            ax_index = 0
            row['quality'] = defaultQ
            for pn, pv in self.plots.items():

                if pn == 'dynamic moduli' and pv:
                    d = self.read_rheo_data(self.dname(row, register['dynamic_moduli']))
                    mw = Maxwell(d, row, dofit=dofit, omega_max=maxwell_omega_max)
                    if dofit:
                        mw.fit()
                        self.set_rowparams(row, mw.params)
                    if doplot:
                        mw.plot(ax[ax_index], **plot_kws)
                    if plot_fit:
                        mw.plot_fit()
                    ax_index += 1

                if pn == 'cole-cole' and pv:
                    d = self.read_rheo_data(self.dname(row, register['dynamic_moduli']))
                    cc = ColeCole(d, row, dofit=dofit, omega_max=cole_omega_max,
                                  gpmax=gpmax)
                    if dofit:
                        cc.fit(fitg0, g0range)
                        self.set_rowparams(row, cc.params)
                    if doplot:
                        cc.plot(ax[ax_index], **plot_kws)
                    if plot_fit:
                        cc.plot_fit(fitg0=fitg0)
                    h_plot.append((index, cc.pl_dat))
                    ax_index += 1

                if pn == 'dynamic viscosity' and pv:
                    d = self.read_rheo_data(self.dname(row, register['dynamic_moduli']))
                    self.plot_dynamic_viscosity(d, ax[ax_index], **plot_kws)
                    ax_index += 1

                if pn == 'temperature' and pv:
                    d = self.read_rheo_data(self.dname(row, register['all']))
                    self.plot_temperature(d, ax[ax_index], **plot_kws)
                    ax_index += 1

                if pn == 'flow curve eta' and pv:
                    d = self.read_rheo_data(self.dname(row, register['flow_curve']))
                    if self.interactive:
                        rng = [x.value for x in self.fc_box.children]
                    else:
                        rng = [0,np.inf, 0, np.inf]
                    eta, deta = self.plot_flow_curve_eta(d, ax[ax_index], rng[:2], rng[2:], **plot_kws)
                    row['eta0'] = eta
                    row['d_eta0'] = deta
                    ax_index += 1

                if pn == 'flow curve tau' and pv:
                    d = self.read_rheo_data(self.dname(row, register['flow_curve']))
                    self.plot_flow_curve_tau(d, ax[ax_index], **plot_kws)
                    ax_index += 1

                # write data to dictionary
                if pv:
                    self.data[pn][index] = d

            self.add_columns(row.index)
            df.loc[index] = row
            # self.table.df.loc[index] = row


        if self.interactive:
            legstr = [str(x[0]) for x in h_plot]
            h = [x[1] for x in h_plot]
            ax[0].legend(h, legstr, bbox_to_anchor=(0, 1.02, 1, 0.2), loc="lower left",
            mode="expand", borderaxespad=0, ncol=len(legstr))
            plt.show()
            self.table.df = df

    def get_gp(self, omega=10, key=None):

        """Get G' and G'' for a fixed frequency.
        """

        gpn = "G' in Pa"
        gppn = "G'' in Pa"
        omn = "omega in rad/s"
        data = self.data['dynamic moduli']
        indices = data.keys()
        npoints = len(indices)

        out = np.zeros((npoints, 3))

        for n, index in enumerate(indices):
            dati = data[index]
            ind = np.abs(dati[omn]-omega).idxmin()
            out[n] = dati.loc[ind, [omn, gpn, gppn]]

            if key is not None:
                out[n,0] = self.df.loc[index, key]

        return out

    def interact(self, ):

        def update_plot(*args):
            for b in togPlot:
                self.plots[b.description] = b.value
            self.init_figure()
            with self.output:
                clear_output(wait=True)
                self.plot(self.table.get_changed_df(),
                              dofit=tb_plotfit.value,
                              cole_omega_max=ft_cc.value,
                              maxwell_omega_max=ft_mw.value,
                              g0range=frs_g0fit.value,
                              fitg0=tb_fitg0.value,
                              defaultQ=ft_quality.value,
                              gpmax=ft_gpmax.value
                              )
                display(self.fig)

            # self.df = df

        def write_df(*args):
            df = self.table.get_changed_df()
            self.df.update(df)
            # self.table.df = self.df
            self.df['plot'] = False

        def save(*args):
            write_df()
            pickle.dump(self.df, open(t_save.value, 'wb'))

        def reset_plot(*args):
            df = self.table.get_changed_df()
            df['plot'] = False
            self.table.df = df

        def update_table(*args):
            self.table.df.update(self.table.get_changed_df())

        self.interactive = True

        self.table = qgrid.show_grid(self.df,
                                       grid_options={
                                        'forceFitColumns': False,
                                        'defaultColumnWidth': 70,
                                       })

        box_layout = Layout(overflow_x='scroll',
                            flex_flow='row',
                            display='flex')
        btable = Box(children=[self.table], layout=box_layout)

        w = widgets.SelectMultiple(
            options=self.table.get_changed_df()['file'],
            value=[],
            # rows=10,
            description='Rheo Data Sets',
            disabled=False,
            layout=widgets.Layout(width='40%', height='160px'),
            style={'description_width': 'initial'}
        )

        items_layout = Layout(flex='1 1 auto', width='auto')

        box_layout = Layout(display='flex',
                            flex_flow='row',
                            width='75%')

        togPlot = [widgets.ToggleButton(description=word,
                                        value=bool(value),
                                        layout=items_layout,
                                        button_style='info')
                   for word, value in self.plots.items()
                   ]

        plotl = Box(children=togPlot, layout=box_layout)

        b_plot = widgets.Button(
            description='plot',
            button_style='info',  # 'success', 'info', 'warning', 'danger' or ''
        )
        b_plot.on_click(update_plot)

        tb_plotfit = widgets.ToggleButton(
            description='plot fit',
            value=True,
            button_style='info'
        )

        b_wpars = widgets.Button(
            description='write fit params',
            button_style='info',
        )
        b_wpars.on_click(write_df)

        t_save = widgets.Text(value='RheoAnalysis.pkl',
                              disabled=False
                              )
        ft_quality = widgets.FloatText(
                        value=1,
                        min=0,
                        max=1,
                        step=0.2,
                        description='Quality:',
                        disabled=False,
                        layout=Layout(width='15%')
                    )

        b_save = widgets.Button(
            description='save',
            button_style='info',
        )
        b_save.on_click(save)

        b_resetplot = widgets.Button(
            description='reset plot',
            button_style='info',
        )
        b_resetplot.on_click(reset_plot)

        b_updatetable = widgets.Button(
            description='update',
            button_style='info',
        )
        b_updatetable.on_click(update_table)

        # Widgets Maxwell Plot
        ft_mw = widgets.FloatText(
            value=6,
            description=r'MW: $\omega_{max}$',
            disabled=False,
            layout=Layout(width='15%')
        )

        # Widgets Cole-Cole Plot
        ft_cc = widgets.FloatText(
            value=50,
            description=r'CC: $\omega_{max}$',
            disabled=False,
            layout=Layout(width='15%')
        )
        ft_gpmax = widgets.FloatText(
            value=90,
            description=r"$G'_{max}$",
            disabled=False,
            layout=Layout(width='15%')
        )

        tb_fitg0 = widgets.ToggleButton(
            description=r'fit $G_0$',
            value=False,
            button_style='info'
        )

        frs_g0fit = widgets.FloatRangeSlider(
            value=[6, 24],
            min=1,
            max=150,
            step=1,
            description=r'$\omega$ range:',
            disabled=False,
            continuous_update=False,
            orientation='horizontal',
            readout=True,
            readout_format='.0f',
        )

        # Widgets Flow Curve
        fls_fc_gpmin = widgets.FloatLogSlider(
            value=0.02,
            base=10,
            min=-2,  # max exponent of base
            max=1,  # min exponent of base
            step=0.2,  # exponent step
            description=r'FC: GP min'
        )
        fls_fc_gpmax = widgets.FloatLogSlider(
            value=0.2,
            base=10,
            min=-2,  # max exponent of base
            max=1,  # min exponent of base
            step=0.2,  # exponent step
            description='GP max'
        )
        fls_fc_etamin = widgets.FloatLogSlider(
            value=1e-3,
            base=10,
            min=-3,  # max exponent of base
            max=1,  # min exponent of base
            step=0.2,  # exponent step
            description=r'FC: eta min'
        )
        fls_fc_etamax = widgets.FloatLogSlider(
            value=100,
            base=10,
            min=-2,  # max exponent of base
            max=2,  # min exponent of base
            step=0.2,  # exponent step
            description='eta max'
        )
        self.cc_box = widgets.HBox([ft_cc, frs_g0fit, tb_fitg0, ft_gpmax])
        self.mw_box = widgets.HBox([ft_mw, ft_quality])
        self.fc_box = widgets.HBox([fls_fc_gpmin, fls_fc_gpmax, fls_fc_etamin, fls_fc_etamax])


        self.output = widgets.Output(layout=Layout(height='600px', width = '800px', border='solid'))
        gui = widgets.VBox([widgets.HBox([b_plot, t_save, b_save, b_resetplot, tb_plotfit, b_wpars,]),
                        plotl, self.mw_box, self.cc_box, self.fc_box] )
        main_window = widgets.VBox([btable, gui, self.output])
        display(main_window)




import numpy as np
from matplotlib import pyplot as plt
import pickle
from mpldatacursor import datacursor
import FitPoissonGamma
import MBetaFormulas
import FitBetaDist
import niceplot


class PhotonProbs(object):
    '''docstring'''
    def __init__(self, run, darks, datdir = './', savdir = './'):
        """Constructor
        """
        self.NValues  = []
        self.Run      = run
        self.Darks    = darks
        self.Prob     = []
        self.Beta     = {'Fit':np.empty((0,4),dtype=np.float32),
                           'Formula':np.empty((0,4),dtype=np.float32)}
        self.Beta_dat = np.empty((0,4),dtype=np.float32)
        self.Filename = 'tmp.pkl'
        self.dt       = 1.
        self.jfunc    = dict()

    def add_probs(self, n, file):
        """Add photon probability matrix to PhotonProbs object.
        """
        probs = np.loadtxt(file)
        if n in self.NValues:
            np.concatenate((self.Prob[self.NValues==n], probs))
        else:
            self.Prob.append(probs)
            self.NValues.append(n)

    def calc_beta(self, n, k, method, disteval='Fit'):
        idx = self.NValues.index(n)
        kb  = self.Prob[idx][:,0]
        nkb = kb.size
        sp  = np.ones((nkb,1),dtype=np.uint8)
        if method=='Fit':
            pb      = self.Prob[idx][:,k+1]
            fit_res = FitPoissonGamma.FitPG(kb, pb, k)
            beta    = fit_res[0][0]**(-1)
            err     = fit_res[1][0]*fit_res[0][0]**(-2)
            pars    = [n,k,beta,err]
        elif method=='Formula':
            if k==12:
                pb    = self.Prob[idx][:,2:4]
                M, dM = MBetaFormulas.formula_M(kb, pb[:,0], pb[:,1])
                tmp   = np.concatenate([sp*n, sp*k, 1./M, 1./M**2*dM], axis=1)
            elif k==0:
                pb = self.Prob[idx][:,1]
                beta0,dbeta0 = MBetaFormulas.formula_beta0(kb, pb)
                tmp = np.concatenate([sp*n, sp*k, beta0, dbeta0], axis=1)
            elif k==1:
                pb = self.Prob[idx][:,1]
                beta1,dbeta1 = MBetaFormulas.formula_beta1(kb,pb)
                tmp = np.concatenate([sp*n,sp*k,beta1,dbeta1],axis=1)
            self.Beta_dat = np.vstack((self.Beta_dat,tmp))
            beta = tmp[:,2]
            if disteval == 'Fit':
                pars_fit = FitBetaDist.fitbetanorm(beta)
                pars = [n, k, pars_fit[0], pars_fit[1]]
            elif disteval == 'mean':
                pars = [n, k, np.mean(beta), np.std(beta)]
        self.Beta[method] = np.vstack((self.Beta[method], pars))

    def plot_poissongamma(self, n, kmax=2):
        idx = self.NValues.index(n)
        kb = self.Prob[idx][:,0]
        pb = self.Prob[idx][:,1:kmax+2]
        for ik in np.arange(0,kmax+1):
            plt.loglog(kb,pb[:,ik],'o')
            idx = (self.Beta['Fit'][:,0]==n)&(self.Beta['Fit'][:,1]==ik)
            plt.loglog(kb,FitPoissonGamma.PoissonGamma(kb,ik,self.Beta['Fit'][idx,2]**(-1)))
        # niceplot.niceplot('plot_poissongamma')
        plt.show()

    def plot_jfunction(self, method, k, errf=1.):
        self.make_jfunctions()
        t = self.jfunc[method + str(k)][:,0]
        beta = self.jfunc[method + str(k)][:,1]
        err = self.jfunc[method + str(k)][:,2]
        pl = plt.errorbar(t, beta, err*errf, marker='o', label='J {}, k={}'.format(method,k))
        plt.legend(loc='best')
        # niceplot.niceplot('plot_jfunction')
        plt.show()

    def make_jfunctions(self):
        for mth in self.Beta:
            ki = np.unique(self.Beta[mth][:,1]).astype(np.uint8)
            for k in ki:
                idx = np.where(self.Beta[mth][:,1]==k)[0]
                bdx = self.Beta[mth][idx,:]
                bdx[:,0] = (bdx[:,0]-1)*self.dt
                bdx = bdx[:,[0, 2, 3]]
                ind = np.argsort(bdx[:,0])
                self.jfunc[mth + str(k)] = bdx[ind,:]

    def set_beta(self,method,n,k,beta,err):
        idx = np.where((self.Beta[method][:,0]==n)&(self.Beta[method][:,1]==k))
        pars = [n,k,beta,err]
        if idx[0].size!=0:
            self.Beta[method][idx] = pars
        else:
            self.Beta[method] = np.vstack((self.Beta[method],pars))

    def load(self,update=True):
        f = open(self.Filename,'rb')
        tmp_dict = pickle.load(f)
        f.close()          
        if update:
            self.__dict__.update(tmp_dict)
        elif not update:
            self.__dict__ = tmp_dict

    def save(self):
        f = open(self.Filename,'wb')
        pickle.dump(self.__dict__,f,2)
        f.close()



#! /usr/bin/env python
from sys import argv,exit,stdout,path
import numpy as np
import os
import string
import matplotlib
from matplotlib import pyplot as plt

def niceplot(ax, kind='default', autoscale=True, grid=True, lfs=10, labelsize=14, ticksize=12):
    if autoscale:
        ax.autoscale(enable=True, axis='both')
    #plt.legend(bbox_to_anchor=(1, 1))#, bbox_transform=plt.gcf().transFigure)
    ax.tick_params(labelsize=ticksize)
    ax.xaxis.label.set_size(labelsize)
    ax.yaxis.label.set_size(labelsize)
    ax.minorticks_on()
    if grid:
        ax.grid(1)
    ax.tick_params('both', length=6, width=1, which='major', direction='out')
    ax.tick_params('both', length=3, width=.5, which='minor', direction='out')
    plt.setp(ax.get_lines(), linewidth=1.5)
    for child in ax.get_children():
        if isinstance(child, matplotlib.spines.Spine):
            child.set_color((.68,.68,.68))
        if isinstance(child, matplotlib.legend.Legend):
            texts = child.get_texts()
            for t in texts:
                t.set_fontsize(lfs)
'''
def get_color(i):
    colors = [(31,119,180),
              (255,127,14),
              (44,160,44),
              (214,39,40),
              (148,103,189),
              (140,86,75),
              (127,127,127),
              (188,189,34),
              (23,190,207),
              (23,190,207),
              (255,128,14),
              (171,171,171),
              (95,158,209),
              (89,89,89),
              (0,107,164)]
    i = i%(len(colors)-1)
#    plt.autoscale(enable=True, axis='both', tight=True)
    colors[:] = [[x/255. for x in y] for y in colors]
    return colors[i]
#    ax.legend(bbox_to_anchor=(0, 1), loc=2)

def set_axes_probs(ax):
    ax.tick_params(labelsize=14)
    ax.minorticks_on()
    ax.tick_params('both', length=8, width=1, which='major', direction='out')
    ax.tick_params('both', length=4, width=.5, which='minor', direction='out')
    ax.xaxis.label.set_fontsize(16)
    ax.yaxis.label.set_fontsize(16)
    plt.tight_layout()
    for child in ax.get_children():
        if isinstance(child, matplotlib.spines.Spine):
            child.set_color((.5,.5,.5))
            child.set_linewidth(8)
      
def niceplot(name):
  fig = plt.gcf()
  fig.set_facecolor('white')
  if name=='plot_poissongamma':
    ax = plt.gca()
    set_axes_probs(ax)
    ax.set_xlabel(r'$\langle k\rangle$')
    ax.set_ylabel(r'$\mathsf{photon probability}$')
    for idx,line in enumerate(ax.get_children()):
      line.set_color(get_color(idx))
      line.set_linewidth(2.0)
    #plt.autoscale(enable=True, axis='x', tight=True)
    ax.set_ylim([1e-7,2.])
  if name=='plot_jfunction':
    ax = plt.gca()
    set_axes_probs(ax)
    ax.set_xlabel(r'$n$')
    ax.set_ylabel(r'$\beta$')
    ax.set_xscale('log')

    #for idx,line in enumerate(ax.get_lines()):
    #  line.set_color(get_color(idx))
    #  line.set_linewidth(2.0)
    #plt.autoscale(enable=True, axis='x', tight=True)
    #ax.set_ylim([1e-7,2.])
'''  



import numpy as np
import lmfit
from fit.BandyFunctions import *
from matplotlib import pyplot as plt
from misc.niceplot import niceplot

def fitbandy( t, cf, err=None, mode='standard', modes=1, init={}, fix=None,
           doplot=False, marker='o', qv=None, h_plot=None, ax=None, output='pars', xl=None, ylim=None,
           color=None):

    #make initial guess for parameters
    for i in range(modes):
        for s in 'tgb':
            vn = s+'{}'.format(i)
            if vn not in init.keys():
                if s == 't':
                    t0 = np.percentile(t,100/(modes+1)*(i+1))
                    init[vn] = (t0, t0/100, t0*100)
                elif s == 'g':
                    init[vn] = (1, .2, 1.8)
                elif s == 'b':
                    init[vn] = (.1, 0, 1)
    if 'a' not in init.keys():
        init['a'] = (0., 0., .2)

    #initialize parameters
    pars = lmfit.Parameters()
    for i in range(modes):
        for s in 'tgb':
            vn = s+'{}'.format(i)
            pars.add(vn, value=init[vn][0], min=init[vn][1], max=init[vn][2], vary=1)
    pars.add('a', value=init['a'][0], min=init['a'][1], max=init['a'][2], vary=0)
    
    if 'off' in mode:
        pars['a'].set(vary=1)
            
    if fix is not None:
        for vn in fix.keys():
            pars[vn].set(value=fix[vn], vary=0)

    if err is not None:
        wgt = 1./err
    else:
        wgt = 1./t
        
    if 'semilogx' in mode:
        if err is not None:
            wgt = 1./np.log10(err)
        else:
            wgt = 1./np.log10(t)
            
    # define residual function
    def res(pars, x, data=None, eps=None):
        """2D Residual function to minimize
        """
        v = pars.valuesdict()
        for i in range(modes):
            if i == 0:
                model = V2a(x, t=v['t{}'.format(i)], b=v['b{}'.format(i)],
                           g=v['g{}'.format(i)], a=v['a'])
            else:
                model += V2a (x, t=v['t{}'.format(i)], b=v['b{}'.format(i)],
                            g=v['g{}'.format(i)], a=0)
        if eps is not None:
            resid = np.abs(data - model)/np.abs(eps)
        else:
            resid = np.abs(data - model)
        return resid

    out = lmfit.minimize(res, pars, args=(t,), kws={'data':cf, 'eps':err}, nan_policy='omit')

    # do all the plotting
    if doplot:
        if xl is None:
            if ax is None:
                ax = plt.gca()
            xl = ax.get_xlim()
            if xl[0] == 0:
                xl = (np.min(t)*0.9,np.max(t)*1.1)
        xf = np.logspace(np.log10(xl[0]),np.log10(xl[1]),50)
        
        for i in range(modes):
            if i == 0:
                g2f = V2a(xf, t=out.params['t{}'.format(i)].value,
                         b=out.params['b{}'.format(i)].value,
                         g=out.params['g{}'.format(i)].value,
                         a=out.params['a'].value)
            else:
                g2f += V2a(xf, t=out.params['t{}'.format(i)].value,
                          b=out.params['b{}'.format(i)].value,
                          g=out.params['g{}'.format(i)].value, a=0)

        if 'legf' in doplot:
            pard = {'t':r'$t: {:.2g}\mathrm{{s}},\,$',
                    'g':r'$\gamma: {:.2g},\,$',
                    'b':r'$\mathrm{{b}}: {:.2g},\,$',
                    'a':r'$\mathrm{{a}}: {:.3g},\,$'}
            labstr_fit = ''
            for i in range(modes):
                for vn in 'tgba':
                    if vn == 'a' and i > 0:
                        continue
                    elif vn == 'a' and i == 0:
                        vnn = 'a'
                    else:
                        vnn = vn + str(i)
                    if vnn in out.var_names:
                        labstr_fit += pard[vn].format(out.params[vnn].value)
                    elif fix is not None and vnn in fix.keys():
                        labstr_fit += 'fix '+ pard[vn].format(fix[vnn])
                    else:
                        labstr_fit += pard[vn].format(0)
        else:
            labstr_fit = ''

        if 'legd' in doplot:
            if qv is not None:
                labstr_data = r'$\mathsf{{q}} = {:.3f}\,\mathsf{{nm}}^{{-1}}$'.format(qv*10)
            else:
                labstr_data = 'data'
        else:
            labstr_data = ''

        pl = []
        if 'fit' in doplot:
            if 'log' in doplot:
                g2f = np.log(np.sqrt(out.best_values['b']/(g2f-1)))
                xf = xf/out.best_values['t']
            pl.append(ax.plot(xf, g2f, '-', label=labstr_fit, linewidth=1))

        if 'data' in doplot:
            if 'log' in doplot:
                t = t/out.best_values['t']
                cf = np.sqrt((cf-1)/out.best_values['b'])
                pl.append(ax.plot(t, cf, marker, markersize=2.5))
            pl.append(ax.plot(t, cf, marker, label=labstr_data, markersize=2.5))
            
        if color is None:
            if h_plot is not None:
                color = h_plot.get_color()
            elif 'data' in doplot:
                color = pl[0][0].get_color()
            else:
                color = 'gray'
                
        if 'log' in doplot:
            ax.set_xscale('linear')
            ax.set_yscale('linear')
        else:
            ax.set_xscale('log')
            ax.set_yscale('linear')

        for p in pl:
            p[0].set_color(color)

        if 'leg' in doplot:
            ax.legend()

        if 'report' in doplot:
            print(out.fit_report())

        ax.set_xlabel(r'delay time $\tau$ [s]')
        ax.set_ylabel(r'$g_2(\tau)$')

        niceplot(ax)
    if output == 'pars':
        pars_arr =  np.zeros((3*modes+1,2))
        for i,vn in enumerate(pars.keys()):
            pars_arr[i,0] = out.params[vn].value
            pars_arr[i,1] = 1.*out.params[vn].stderr
        return pars_arr
    elif output == 'fit':
        return out



import numpy as np
import multiprocessing as mp
from queue import PriorityQueue
from multiprocessing.managers import SyncManager
from ..helper import *
import time
import h5py as h5
from ..misc.progressbar import progress


class MyManager(SyncManager):
    pass

def Manager():
    m = MyManager()
    m.start()
    return m

def init_h5_file():
    pass
    

def h5_dump(fname, nframes, dataQ, dtype=np.dtype('uint16')):

    tcalc = time.time()

    with h5.File(fname, 'w', driver='stdio') as f:
        t0 = 0
        last_chunk = -1
        while t0 < nframes:
            chunk = dataQ.get()
            chunk_size = chunk[1].shape[0]
            if t0 == 0:
                f.require_dataset('data', (nframes,*chunk[1].shape[1:]), dtype=dtype, compression='gzip')
            chunk_diff = chunk[0] - last_chunk
            if chunk_diff != 1:
                raise IOError('Chunks have been read in wrong order: chunk index difference is % and not 1.' %
                                 chunk_diff)
            last_chunk, f['/data'][t0:t0+chunk_size,...] = chunk
            t0 += chunk_size
            progress(t0,nframes)
        progress(1,1)

    #END OF MAIN LOOP put results to output queue
    tcalc = time.time() - tcalc
    print('Elapsed time: %.2fs' % tcalc)

def h5_update_meta(sid, meta, fname):
    
    with h5.File(fname, 'r+', driver='stdio') as f:
        for att in meta.columns:
            f['/data'].attrs[att] = meta.loc[sid, att]


def to_h5(obj, sid, filename, nprocs=4, chunk_size=100, dtype=np.dtype('uint16'), **kwargs):

    print('\n\n#### Writing Data to HDF5 File ####\nSeries: {} in folder {}\n'.format(sid, obj.meta.loc[sid, 'datdir']))

    read_data_opt = {'dtype':dtype,
                     'nprocs':1,
                     'verbose':False,
    }
    read_data_opt.update(kwargs)

    fmax = obj.meta.loc[sid,'nframes']
    chunks = [np.arange(i*chunk_size,min((i+1)*chunk_size,fmax))
              for i in range(np.ceil(fmax/chunk_size).astype(np.int32))]

    print('Using {} processes to read data.'.format(nprocs))
    MyManager.register("PriorityQueue", PriorityQueue)  # Register a shared PriorityQueue
    m = Manager()
    dataQ = m.PriorityQueue(nprocs)
    indxQ = m.PriorityQueue()
    #dataQ = mp.Queue(nread_procs)
    #indxQ = mp.Queue()
    read_data_opt['dataQ'] = dataQ
    read_data_opt['indxQ'] = indxQ
    read_data_opt['method'] = 'queue_chunk'

    for i,chunk in enumerate(chunks):
        indxQ.put((i,chunk))

    lock = 0
    if 'h5' in obj.fmtstr:
        lock = mp.Lock()
        read_data_opt['lock'] = lock

    procs = []
    for ip in range(nprocs):
        procs.append(mp.Process(target=obj.get_series,
                                args=(sid,), kwargs=read_data_opt))
        procs[ip].start()
        if ip == 0:
            time.sleep(2)

    h5_dump(filename, fmax, dataQ)
    h5_update_meta(sid, obj.meta, filename)

    dataQ.task_done()
    indxQ.task_done()

    # stopping processes
    for ip in range(nprocs):
        procs[ip].join()

    # closing queues
    #dataQ.close()
    #dataQ.join_thread()
    #indxQ.close()
    #indxQ.join_thread()

    print('Dataset saved to %s.' % filename)



import numpy as np
from matplotlib import pyplot as plt
import matplotlib


class AnaList:
    
    def __init__(self, Xana, cmap='magma'):

        self.Xana = Xana
        self.colors = None
        self.cmap = cmap
        self.update_colors()
        self.markers = ['o',]
        
    def __str__(self):
        return 'Analist: super class for analysis classes.'

    def __repr__(self):
        return self.__str__()

    def update_colors(self, cmap=None, factor=1, repeat=1,):
        """Return vector of colors for plots
        """
        factor = int(factor)
        if cmap is None:
            cmap = self.cmap
        else:
            self.cmap = cmap

        self.colors = []
        cm = plt.get_cmap(cmap, factor)
        ci = np.linspace(0,1,factor)
        for i in range(factor):
            self.colors.append(cm(i))
        self.colors = np.tile(self.colors, (repeat, 1))

    def update_markers(self, nmarkers, change_marker=0):
        """Return vector of markers for plots
        """
        self.markers = ['o',]
        if change_marker:
            mlist = list(matplotlib.markers.MarkerStyle.markers.keys())[:-4]
            del mlist[2], mlist[1]
            self.markers.extend(mlist)
        else:
            self.markers = ['o',]*nmarkers
    

    
   






def progress(count, total, suffix=''):
    bar_len = 60
    filled_len = int(round(bar_len * count / float(total)))

    percents = round(100.0 * count / float(total), 1)
    bar = '=' * filled_len + '-' * (bar_len - filled_len)
    if percents < 100:
        print('\r[{}] {}{} ...{}'.format(bar, percents, '%', suffix), end='', flush=1)
    else:
        print('\r[{}] {}{}'.format(bar, percents, '%'), end='\n', flush=1)


import numpy as np
from matplotlib import pyplot as plt
from matplotlib.colors import LogNorm
from ipywidgets import interact, fixed

from readDet.openFile import getNXimagedata
from readDet.processData import getFilenames
from misc.nxsinfo import nxsinfo

import h5py

def getimg(run, imgn, fnames, dim, ffmask=0, qroi=0, rgn=0):
    img = getNXimagedata(fnames[run], first=(imgn,0,0), last=(imgn+1,dim[1],dim[0]))
    img = np.squeeze(img)
    if rgn:
        img = img[rgn[0]:rgn[1],rgn[2]:rgn[3]]
        if type(ffmask) == np.ndarray:
            ffmask = ffmask[rgn[0]:rgn[1],rgn[2]:rgn[3]]
    if type(ffmask) == np.ndarray:
        img = img * ffmask
        gp = np.sum(ffmask>0)
    else:
        gp = dim[0]*dim[1]
    if qroi:
        for i in range(len(qroi)):
            img[qroi[i]] = np.max(img)
    return img, gp

def browse_images(nf=0, run=0, fnames=0, dim=0, exp=0, gp=0, im=0, txt=0, fig=0, ffmask=0, qroi=0, rgn=0):
    def view_image(i=0, **kargs):
        img = getimg(run, i, fnames, dim, ffmask, qroi, rgn)[0]
        im.set_data(img)
        textstr = get_text(img, run, i, exp, gp)
        txt.set_text(textstr)
        fig.canvas.draw_idle()
    interact(view_image, i=(0, nf-1), run=fixed(run), fnames=fixed(fnames), dim=fixed(dim), exp=fixed(exp),
                 gp=fixed(gp), im=fixed(im), txt=fixed(txt), fig=fixed(fig), ffmask=fixed(ffmask), qroi=fixed(qroi),
                 rgn=fixed(rgn))
    
def get_text(img, run, imgn, exp, gp):
    img_sum = np.sum(img)
    img_mean = img_sum / gp
    mx = np.max(img)
    if type(exp) == np.ndarray:
        exp = exp[run]
        ill = imgn*(exp+0.001)
    textstr = (r'$\mathrm{{I}}_{{img}} = {0:.3g}$'+'\n'+ 
           r'$\langle \mathrm{{I}}\rangle _{{pix}} = {1:.3g}$'+'\n'+
           r'$\mathrm{{I}}_{{max}} = {6:.3g}$'+'\n'+
           r'$\mathrm{{t}}_{{exp}} = {2:.3g}s$'+'\n'+ 
           r'$\mathrm{{t}}_{{ill}} = {3:.3g}s$'+'\n'+ 
           r'$\mathrm{{run}}={4:d}$'+'\n'+ 
           r'$\mathrm{{imgn}}={5:d}$').format(img_sum, img_mean, exp, ill, run, imgn, mx)
    return textstr

def plot1img(run, datdir, imgn, ffmask=0, rgn=0, exp=0, qroi=0, cb=0, norm=None):
    if type(datdir) == str:
        fnames = getFilenames(datdir, 'nxs')
    else:
        fnames = datdir
    ct,nf = nxsinfo(fnames[run])
    if type(ffmask) == np.ndarray:
        m,n = np.shape(ffmask)
    else:
        m,n = np.shape(getNXimagedata(fnames[run], last=(1,)))[1:]
    img, gp = getimg(run, imgn, fnames, (m, n), ffmask, qroi, rgn)
    out = img.copy()
    fig, ax = plt.subplots(figsize=(9,6))
    ax.set_xticklabels('')
    ax.set_yticklabels('')
    ax.tick_params('both', length=0, width=0, which='major')
    if cb:
        im = ax.imshow(img, vmin=cb[0], vmax=cb[1], cmap=plt.get_cmap('magma'), norm=norm)
    else:
        im = ax.imshow(img, cmap=plt.get_cmap('magma'), norm=norm)
    plt.colorbar(im)      
    textstr = get_text(img, run, imgn, exp, gp)
    props = dict(boxstyle='round', facecolor='white')
    txt = ax.text(-0.4, 0.98, textstr, transform=ax.transAxes, fontsize=14,
            verticalalignment='top',horizontalalignment='left', bbox=props)
    propdic = {'nf':nf, 'run':run, 'fnames':fnames, 'dim':(m, n), 'exp':exp, 'gp':gp, 'im':im, 'txt':txt, 'fig':fig,
             'ffmask':ffmask, 'qroi':qroi, 'rgn':rgn}
    plt.show(block=0)
    return out, propdic



import numpy as np

def betaratio(kv, kb, prob, err=None, perframe=True):
    """Calculate speckle contrast from photon probability ratios.
    """
    prob = np.ma.array(prob, mask=np.zeros_like(prob, dtype=np.bool))
    kb = np.ma.array(kb, mask=kb==0)
    beta = np.ma.array(np.zeros((kv.size-1,kb.size)))
    dbeta = beta.copy()
    
    for i,ki in enumerate(kv[:-1]):
        c = i+2 < kv.size
        prob[i:i+1+c][prob[i+1:i+2+c]==0] = np.ma.masked
        p1 = prob[i]
        p2 = prob[i+1]
        a = p1 / p2
        divd = a*kb - (ki + 1.)
        divs = 1. + (1.-a) * ki
        beta[i] = divd/kb/divs
        prob.mask = np.ma.nomask

    if err is not None:
        dbeta = 1./kb*np.sqrt(2*(1+np.abs(beta))/(err[0]*err[1]))
    else:
        dbeta = np.ones((kb.size,1), dtype=np.float32)

    return beta, dbeta



import numpy as np
from matplotlib import pyplot as plt
from lmfit.models import GaussianModel as gauss_mod
from lmfit.models import VoigtModel as voigt_mod
from lmfit.models import LinearModel as lin_mod

def find_center(img, mask=None, doplot=False, fit_report=False):
    
    mod = voigt_mod() + lin_mod()
    if mask is None:
        mask = np.ones_like(img).astype('bool')
        
    ctr = np.zeros(2)
    
    if doplot:
        fig, ax = plt.subplots(2,1)
    
    for i in range(2):
        dat = np.ma.masked_array(img, mask=~(mask).astype('bool'))
        dat = dat.mean(i)
        par = mod.make_params()
        x = np.arange(len(dat))
        par['center'].set(np.mean(dat*x)/dat.mean())
        par['slope'].set(0, vary=False)
        par['intercept'].set(dat.mean())
        res = mod.fit(dat, params=par, x=x)
        
        if doplot:
            res.plot_fit(ax=ax.flat[i])
            if i:
                ax.flat[i].set_title('')
        if fit_report:
            print(res.fit_report())
            
        ctr[i] = res.params['center'].value
    return ctr.round().astype('int16')



import numpy as np
import pyFAI

class Agipd1m(pyFAI.detectors.Detector):

    def __init__(self, *args, **kwargs):

        super().__init__(pixel1=2e-4, pixel2=2e-4, max_shape=(8192,128),)
        self.shape = self.dim = (8192,128) 
        self.aliases = ['Agipd1m']
        self.IS_CONTIGUOUS = False
        self.mask = np.zeros(self.shape)

        # self._pixel_corners = spatial_distortion.astype(np.float32)



#/*##########################################################################
# Copyright (C) 2004-2007 European Synchrotron Radiation Facility
#
# This file is part of the PyMCA X-ray Fluorescence Toolkit developed at
# the ESRF by the Beamline Instrumentation Software Support (BLISS) group.
#
# This toolkit is free software; you can redistribute it and/or modify it 
# under the terms of the GNU General Public License as published by the Free
# Software Foundation; either version 2 of the License, or (at your option) 
# any later version.
#
# PyMCA is distributed in the hope that it will be useful, but WITHOUT ANY
# WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS
# FOR A PARTICULAR PURPOSE.  See the GNU General Public License for more
# details.
#
# You should have received a copy of the GNU General Public License along with
# PyMCA; if not, write to the Free Software Foundation, Inc., 59 Temple Place,
# Suite 330, Boston, MA 02111-1307, USA.
#
# PyMCA follows the dual licensing model of Trolltech's Qt and Riverbank's PyQt
# and cannot be used as a free plugin for a non-free program. 
#
# Please contact the ESRF industrial unit (industry@esrf.fr) if this license 
# is a problem to you.
#############################################################################*/
"""
    EdfFile.py
    Generic class for Edf files manipulation.    

    Interface:
    ===========================
    class EdfFile:          
        __init__(self,FileName)	
        GetNumImages(self)
        def GetData(self,Index, DataType="",Pos=None,Size=None):
        GetPixel(self,Index,Position)
        GetHeader(self,Index)
        GetStaticHeader(self,Index)
        WriteImage (self,Header,Data,Append=1,DataType="",WriteAsUnsigened=0,ByteOrder="")


    Edf format assumptions:
    ===========================
    The following details were assumed for this implementation:
    - Each Edf file contains a certain number of data blocks.
    - Each data block represents data stored in an one, two or three-dimensional array.
    - Each data block contains a header section, written in ASCII, and a data section of
      binary information.
    - The size of the header section in bytes is a multiple of 1024. The header is
      padded with spaces (0x20). If the header is not padded to a multiple of 1024,
      the file is recognized, but the output is always made in this format.
    - The header section starts by '{' and finishes by '}'. It is composed by several
      pairs 'keyword = value;'. The keywords are case insensitive, but the values are case
      sensitive. Each pair is put in a new line (they are separeted by 0x0A). In the
      end of each line, a semicolon (;) separes the pair of a comment, not interpreted.
      Exemple:
        {
        ; Exemple Header
        HeaderID = EH:000001:000000:000000    ; automatically generated
        ByteOrder = LowByteFirst              ;
        DataType = FloatValue                 ; 4 bytes per pixel
        Size = 4000000                        ; size of data section
        Dim_1= 1000                           ; x coordinates
        Dim_2 = 1000                          ; y coordinates

        (padded with spaces to complete 1024 bytes)
        }
    - There are some fields in the header that are required for this implementation. If any of
      these is missing, or inconsistent, it will be generated an error:
        Size: Represents size of data block
        Dim_1: size of x coordinates (Dim_2 for 2-dimentional images, and also Dim_3 for 3d)
        DataType
        ByteOrder
    - For the written images, these fields are automatically genereted:
        Size,Dim_1 (Dim_2 and Dim_3, if necessary), Byte Order, DataType, HeaderID and Image
      These fields are called here "static header", and can be retrieved by the method
      GetStaticHeader. Other header components are taken by GetHeader. Both methods returns
      a dictionary in which the key is the keyword of the pair. When writting an image through
      WriteImage method, the Header parameter should not contain the static header information,
      which is automatically generated.
    - The indexing of images through these functions is based just on the 0-based position in
      the file, the header items HeaderID and Image are not considered for referencing the
      images.
    - The data section contais a number of bytes equal to the value of Size keyword. Data
      section is going to be translated into an 1D, 2D or 3D Numpy Array, and accessed
      through GetData method call.      
"""
__author__ =  'Alexandre Gobbo (gobbo@esrf.fr)'
__version__=  '$Revision: 1.21 $'

################################################################################  
import sys
import re
import numpy
import os.path , tempfile, shutil
import gzip

try:
    from FastEdf import extended_fread
    CAN_USE_FASTEDF = 1
except:
    CAN_USE_FASTEDF = 0

################################################################################
# constants
HEADER_BLOCK_SIZE = 1024
STATIC_HEADER_ELEMENTS=("HeaderID","Image","ByteOrder","DataType",
                        "Dim_1","Dim_2","Dim_3",
                        "Offset_1","Offset_2","Offset_3",
                        "Size")
STATIC_HEADER_ELEMENTS_CAPS=("HEADERID","IMAGE","BYTEORDER","DATATYPE",
                             "DIM_1","DIM_2","DIM_3",
                             "OFFSET_1","OFFSET_2","OFFSET_3",
                             "SIZE")

LOWER_CASE=0
UPPER_CASE=1

KEYS=1
VALUES=2

###############################################################################
class Image:
    """
    """
    def __init__(self):
        """ Constructor
        """
        self.Header={}
        self.StaticHeader={}
        self.HeaderPosition=0
        self.DataPosition=0
        self.Size=0
        self.NumDim=1
        self.Dim1=0
        self.Dim2=0
        self.Dim3=0
        self.DataType=""
        #for i in STATIC_HEADER_ELEMENTS: self.StaticHeader[i]=""
   
################################################################################

class  EdfFile:    
    """
    """
    ############################################################################
    #Interface
    def __init__(self,FileNameOrFileDescriptor,fastedf=None):
        """ Constructor
            FileNameOrFileDescriptor:   Name of the file (either existing or to be created) or file descriptor
        """
        self.Images=[]
        self.NumImages=0

        if isinstance(FileNameOrFileDescriptor,str) :
            self.File = None
            self.FileName = FileNameOrFileDescriptor
        else:
            self.File = FileNameOrFileDescriptor
            try:
                self.FileName = FileNameOrFileDescriptor.name
            except AttributeError:
                self.FileName = FileNameOrFileDescriptor.filename
            
        if fastedf is None:fastedf=0
        self.fastedf=fastedf
        
        if sys.byteorder=="big": self.SysByteOrder="HighByteFirst"
        else: self.SysByteOrder="LowByteFirst"

        if self.File is None:
            self.__ownedOpen = True
            try:
                if os.path.isfile(self.FileName)==0:                
                    self.File = open(self.FileName, "wb")
                    self.File.close()    

                if (os.access(self.FileName,os.W_OK)):
                    self.File=open(self.FileName, "r+b")
                else : 
                    self.File=open(self.FileName, "rb")

                self.File.seek(0, 0)
            except:
                try:
                    self.File.close()
                except:
                    pass
                raise IOError("EdfFile: Error opening file")
        else: self.__ownedOpen = False

        Index=0

        self.__checkEdfFile = re.compile(r'^dim_1',re.IGNORECASE|re.MULTILINE)

        self.File.seek(0, 0)
        if self.File.read(1) == '':
            return 
        self.File.seek(0, 0)

        startHeaderPos = self.File.tell()
        header,headersize = self.__readOneHeader()
        while 1 :
            img = Image()
            img.HeaderPosition = startHeaderPos
            img.DataPosition = startHeaderPos + headersize
            StaticHeader = {}
            for line in header.split('\n') :
                posSemiColumn = line.rfind(';')
                if posSemiColumn < 0: continue
                else:
                    line = line[:posSemiColumn]
                try:
                    key,value = line.split('=')
                    key = key.strip()
                    value = value.strip()
                except ValueError: continue
                keyCap = key.upper()
                if keyCap in STATIC_HEADER_ELEMENTS_CAPS :
                    img.StaticHeader[key] = value
                    StaticHeader[keyCap] = value
                else:
                    img.Header[key] = value
            #Check Image Size
            try:
                stringSize = StaticHeader['SIZE']
                imageSize = int(stringSize)
                img.Size = imageSize
                self.File.seek(img.DataPosition + imageSize)
                startHeaderPos = self.File.tell()
            except KeyError:
                raise KeyError("EdfFile: Image doesn't have size information")
            except ValueError:
                raise ValueError("EdfFile: Image size information is not an integer")
            #Check DataType
            try:
                img.DataType = StaticHeader["DATATYPE"]
            except KeyError:
                raise KeyError("EdfFile: Image doesn't have datatype information")
            #Check ByteOrder
            try:
                img.ByteOrder = StaticHeader["BYTEORDER"]
            except KeyError:
                raise KeyError("EdfFile: Image doesn't have byteorder information")
            #Check DIM1
            try:
                stringDim1 = StaticHeader["DIM_1"]
                img.Dim1 = int(stringDim1)
                try:
                    img.Offset1 = int(StaticHeader.get('OFFSET_1','0'))
                except ValueError:
                    img.Offset1 = 0
            except KeyError:
                raise KeyError("EdfFile: Image doesn't have dimension information")
            #DIM2
            stringDim2 = StaticHeader.get("DIM_2",None)
            if stringDim2 is not None:
                img.Dim2 = int(stringDim2)
                try:
                    img.Offset2 = int(StaticHeader.get("OFFSET_2","0"))
                except ValueError:
                    img.Offset2 = 0
                img.NumDim = 2
            #DIM3
            stringDim3 = StaticHeader.get("DIM_3",None)
            if stringDim3 is not None:
                img.Dim3 = int(stringDim3)
                try:
                    img.Offset3 = int(StaticHeader.get("OFFSET_3","0"))
                except ValueError:
                    img.Offset3 = 0
            #APPEND IMAGE
            self.Images.append(img)
            self.NumImages += 1
            try:
                header,headersize = self.__readOneHeader()
            except: break
        
    def GetNumImages(self):
        """ Returns number of images of the object (and associated file)
        """
        return self.NumImages

        
    
    def GetData(self,Index, DataType="",Pos=None,Size=None):
        """ Returns numpy array with image data
            Index:          The zero-based index of the image in the file
            DataType:       The edf type of the array to be returnd
                            If ommited, it is used the default one for the type
                            indicated in the image header
                            Attention to the absence of UnsignedShort,
                            UnsignedInteger and UnsignedLong types in
                            Numpy Python
                            Default relation between Edf types and NumPy's typecodes:
                                SignedByte          int8   b
                                UnsignedByte        uint8  B       
                                SignedShort         int16  h
                                UnsignedShort       uint16 H
                                SignedInteger       int32  i
                                UnsignedInteger     uint32 I
                                SignedLong          int32  i
                                UnsignedLong        uint32 I
                                Signed64            int64  (l in 64bit, q in 32 bit)
                                Unsigned64          uint64 (L in 64bit, Q in 32 bit) 
                                FloatValue          float32 f
                                DoubleValue         float64 d
            Pos:            Tuple (x) or (x,y) or (x,y,z) that indicates the begining
                            of data to be read. If ommited, set to the origin (0),
                            (0,0) or (0,0,0)
            Size:           Tuple, size of the data to be returned as x) or (x,y) or
                            (x,y,z) if ommited, is the distance from Pos to the end.

            If Pos and Size not mentioned, returns the whole data.                         
        """
        fastedf = self.fastedf
        if Index < 0 or Index >= self.NumImages: raise ValueError("EdfFile: Index out of limit")
        if fastedf is None: fastedf = 0
        if Pos is None and Size is None:
            self.File.seek(self.Images[Index].DataPosition,0)
            datatype = self.__GetDefaultNumpyType__(self.Images[Index].DataType, index= Index)
            try:
                datasize = self.__GetSizeNumpyType__(datatype)
            except TypeError:
                print("What is the meaning of this error?")
                datasize = 8
            if self.Images[Index].NumDim==3:
                sizeToRead = self.Images[Index].Dim1 * \
                             self.Images[Index].Dim2 * \
                             self.Images[Index].Dim3 * datasize
                Data = numpy.fromstring(self.File.read(sizeToRead),
                            datatype)
                Data = numpy.reshape(Data, (self.Images[Index].Dim3,self.Images[Index].Dim2, self.Images[Index].Dim1))
            elif self.Images[Index].NumDim==2:
                sizeToRead = self.Images[Index].Dim1 * \
                             self.Images[Index].Dim2 * datasize
                Data = numpy.fromstring(self.File.read(sizeToRead),
                            datatype)
                #print "datatype = ",datatype
                #print "Data.type = ", Data.dtype.char
                #print "self.Images[Index].DataType ", self.Images[Index].DataType
                #print "Data.shape",Data.shape
                #print "datasize = ",datasize
                #print "sizeToRead ",sizeToRead 
                #print "lenData = ", len(Data)
                Data = numpy.reshape(Data, (self.Images[Index].Dim2, self.Images[Index].Dim1))
            elif self.Images[Index].NumDim==1:
                sizeToRead = self.Images[Index].Dim1 * datasize
                Data = numpy.fromstring(self.File.read(sizeToRead),
                            datatype)
        elif fastedf and CAN_USE_FASTEDF:
            type= self.__GetDefaultNumpyType__(self.Images[Index].DataType, index= Index)
            size_pixel=self.__GetSizeNumpyType__(type)
            Data=Numpy.array([],type)
            if self.Images[Index].NumDim==1:
                if Pos==None: Pos=(0,)
                if Size==None: Size=(0,)
                sizex=self.Images[Index].Dim1
                Size=list(Size)                
                if Size[0]==0:Size[0]=sizex-Pos[0]
                self.File.seek((Pos[0]*size_pixel)+self.Images[Index].DataPosition,0)
                Data = numpy.fromstring(self.File.read(Size[0]*size_pixel), type)
            elif self.Images[Index].NumDim==2:                
                if Pos==None: Pos=(0,0)
                if Size==None: Size=(0,0)
                Size=list(Size)
                sizex,sizey=self.Images[Index].Dim1,self.Images[Index].Dim2
                if Size[0]==0:Size[0]=sizex-Pos[0]
                if Size[1]==0:Size[1]=sizey-Pos[1]
                Data=numpy.zeros([Size[1],Size[0]],type)
                self.File.seek((((Pos[1]*sizex)+Pos[0])*size_pixel)+self.Images[Index].DataPosition,0)
                extended_fread(Data, Size[0]*size_pixel , numpy.array([Size[1]]),
                               numpy.array([sizex*size_pixel]) ,self.File)

            elif self.Images[Index].NumDim==3:
                if Pos==None: Pos=(0,0,0)
                if Size==None: Size=(0,0,0)
                Size=list(Size)
                sizex,sizey,sizez=self.Images[Index].Dim1,self.Images[Index].Dim2,self.Images[Index].Dim3
                if Size[0]==0:Size[0]=sizex-Pos[0]
                if Size[1]==0:Size[1]=sizey-Pos[1]
                if Size[2]==0:Size[2]=sizez-Pos[2]
                Data=numpy.zeros([Size[2],Size[1],Size[0]],type)
                self.File.seek(((((Pos[2]*sizey+Pos[1])*sizex)+Pos[0])*size_pixel)+self.Images[Index].DataPosition,0)
                extended_fread(Data, Size[0]*size_pixel , numpy.array([Size[2],Size[1]]),
                        numpy.array([ sizey*sizex*size_pixel , sizex*size_pixel]) ,self.File)

        else:
            if fastedf:print("I could not use fast routines")
            type= self.__GetDefaultNumpyType__(self.Images[Index].DataType, index= Index)
            size_pixel=self.__GetSizeNumpyType__(type)
            Data=numpy.array([],type)
            if self.Images[Index].NumDim==1:
                if Pos==None: Pos=(0,)
                if Size==None: Size=(0,)
                sizex=self.Images[Index].Dim1
                Size=list(Size)                
                if Size[0]==0:Size[0]=sizex-Pos[0]
                self.File.seek((Pos[0]*size_pixel)+self.Images[Index].DataPosition,0)
                Data = numpy.fromstring(self.File.read(Size[0]*size_pixel), type)
            elif self.Images[Index].NumDim==2:                
                if Pos==None: Pos=(0,0)
                if Size==None: Size=(0,0)
                Size=list(Size)
                sizex,sizey=self.Images[Index].Dim1,self.Images[Index].Dim2
                if Size[0]==0:Size[0]=sizex-Pos[0]
                if Size[1]==0:Size[1]=sizey-Pos[1]
                #print len(range(Pos[1],Pos[1]+Size[1])), "LECTURES OF ", Size[0], "POINTS"
                #print "sizex = ", sizex, "sizey = ", sizey
                Data = numpy.zeros((Size[1],Size[0]), type)
                dataindex =0
                for y in range(Pos[1],Pos[1]+Size[1]):
                    self.File.seek((((y*sizex)+Pos[0])*size_pixel)+self.Images[Index].DataPosition,0)
                    line = numpy.fromstring(self.File.read(Size[0]*size_pixel), type)
                    Data[dataindex,:] =  line[:]
                    #Data=numpy.concatenate((Data,line))
                    dataindex += 1
                #print "DataSize = ",Data.shape
                #print "Requested reshape = ",Size[1],'x',Size[0]
                #Data = numpy.reshape(Data, (Size[1],Size[0]))                            
            elif self.Images[Index].NumDim==3:
                if Pos==None: Pos=(0,0,0)
                if Size==None: Size=(0,0,0)
                Size=list(Size)
                sizex,sizey,sizez=self.Images[Index].Dim1,self.Images[Index].Dim2,self.Images[Index].Dim3
                if Size[0]==0:Size[0]=sizex-Pos[0]
                if Size[1]==0:Size[1]=sizey-Pos[1]
                if Size[2]==0:Size[2]=sizez-Pos[2]
                for z in range(Pos[2],Pos[2]+Size[2]):
                    for y in range(Pos[1],Pos[1]+Size[1]):
                        self.File.seek(((((z*sizey+y)*sizex)+Pos[0])*size_pixel)+self.Images[Index].DataPosition,0)
                        line = numpy.fromstring(self.File.read(Size[0]*size_pixel), type)
                        Data=numpy.concatenate((Data,line))                
                Data = numpy.reshape(Data, (Size[2],Size[1],Size[0]))

        if str.upper(self.SysByteOrder)!=str.upper(self.Images[Index].ByteOrder):
            Data=Data.byteswap()
        if DataType != "":
            Data=self.__SetDataType__ (Data,DataType)
        return Data



    def GetPixel(self,Index, Position):
        """ Returns double value of the pixel, regardless the format of the array
            Index:      The zero-based index of the image in the file
            Position:   Tuple with the coordinete (x), (x,y) or (x,y,z)
        """
        if Index < 0 or Index >= self.NumImages: raise ValueError("EdfFile: Index out of limit")
        if len(Position)!= self.Images[Index].NumDim: raise ValueError("EdfFile: coordinate with wrong dimension ")
        
        size_pixel=self.__GetSizeNumpyType__(self.__GetDefaultNumpyType__(self.Images[Index].DataType), index= Index)
        offset=Position[0]*size_pixel
        if self.Images[Index].NumDim>1:
            size_row=size_pixel * self.Images[Index].Dim1
            offset=offset+ (Position[1]* size_row)
            if self.Images[Index].NumDim==3:
                size_img=size_row * self.Images[Index].Dim2
                offset=offset+ (Position[2]* size_img)
        self.File.seek(self.Images[Index].DataPosition + offset,0)
        Data = Numpy.fromstring(self.File.read(size_pixel), self.__GetDefaultNumpyType__(self.Images[Index].DataType, index= Index))
        if str.upper(self.SysByteOrder)!=str.upper(self.Images[Index].ByteOrder):
            Data=Data.byteswap() 
        Data=self.__SetDataType__ (Data,"DoubleValue")
        return Data[0]
         
        
    def GetHeader(self,Index):
        """ Returns dictionary with image header fields.
            Does not include the basic fields (static) defined by data shape, 
            type and file position. These are get with GetStaticHeader
            method.
            Index:          The zero-based index of the image in the file
        """
        if Index < 0 or Index >= self.NumImages: raise ValueError("Index out of limit")
        #return self.Images[Index].Header
        ret={}
        for i in list(self.Images[Index].Header.keys()):
            ret[i]=self.Images[Index].Header[i]        
        return ret

        
    def GetStaticHeader(self,Index):
        """ Returns dictionary with static parameters
            Data format and file position dependent information
            (dim1,dim2,size,datatype,byteorder,headerId,Image)
            Index:          The zero-based index of the image in the file
        """ 
        if Index < 0 or Index >= self.NumImages: raise ValueError("Index out of limit")
        #return self.Images[Index].StaticHeader
        ret={}
        for i in list(self.Images[Index].StaticHeader.keys()):
            ret[i]=self.Images[Index].StaticHeader[i]        
        return ret


    def WriteImage (self,Header,Data,Append=1,DataType="",ByteOrder=""):
        """ Writes image to the file. 
            Header:         Dictionary containing the non-static header
                            information (static information is generated
                            according to position of image and data format
            Append:         If equals to 0, overwrites the file. Otherwise, appends
                            to the end of the file
            DataType:       The data type to be saved to the file:
                                SignedByte          
                                UnsignedByte               
                                SignedShort         
                                UnsignedShort       
                                SignedInteger       
                                UnsignedInteger     
                                SignedLong          
                                UnsignedLong        
                                FloatValue          
                                DoubleValue         
                            Default: according to Data array typecode:
                                    1:  SignedByte
                                    b:  UnsignedByte
                                    s:  SignedShort       
				    w:  UnsignedShort
                                    i:  SignedInteger
                                    l:  SignedLong          
				    u:  UnsignedLong
                                    f:  FloatValue       
                                    d:  DoubleValue
            ByteOrder:      Byte order of the data in file:
                                HighByteFirst
                                LowByteFirst
                            Default: system's byte order
        """
        if Append==0:
            self.File.truncate(0)
            self.Images=[]
            self.NumImages=0
        Index=self.NumImages
        self.NumImages = self.NumImages + 1                
        self.Images.append(Image())

        #self.Images[Index].StaticHeader["Dim_1"] = "%d" % Data.shape[1]
        #self.Images[Index].StaticHeader["Dim_2"] = "%d" % Data.shape[0]
        if len(Data.shape)==1:
            self.Images[Index].Dim1=Data.shape[0]
            self.Images[Index].StaticHeader["Dim_1"] = "%d" % self.Images[Index].Dim1
            self.Images[Index].Size=(Data.shape[0]*\
                                     self.__GetSizeNumpyType__(Data.dtype))
        elif len(Data.shape)==2:
            self.Images[Index].Dim1=Data.shape[1]
            self.Images[Index].Dim2=Data.shape[0]
            self.Images[Index].StaticHeader["Dim_1"] = "%d" % self.Images[Index].Dim1
            self.Images[Index].StaticHeader["Dim_2"] = "%d" % self.Images[Index].Dim2
            self.Images[Index].Size=(Data.shape[0]*Data.shape[1]*\
                                     self.__GetSizeNumpyType__(Data.dtype))
            self.Images[Index].NumDim=2
        elif len(Data.shape)==3:
            self.Images[Index].Dim1=Data.shape[2]
            self.Images[Index].Dim2=Data.shape[1]
            self.Images[Index].Dim3=Data.shape[0]
            self.Images[Index].StaticHeader["Dim_1"] = "%d" % self.Images[Index].Dim1
            self.Images[Index].StaticHeader["Dim_2"] = "%d" % self.Images[Index].Dim2
            self.Images[Index].StaticHeader["Dim_3"] = "%d" % self.Images[Index].Dim3
            self.Images[Index].Size=(Data.shape[0]*Data.shape[1]*Data.shape[2]*\
                                     self.__GetSizeNumpyType__(Data.dtype))
            self.Images[Index].NumDim=3
        elif len(Data.shape)>3:
            raise TypeError("EdfFile: Data dimension not suported")
        

        if DataType=="":
            self.Images[Index].DataType=self.__GetDefaultEdfType__(Data.dtype)
        else:
            self.Images[Index].DataType=DataType
            Data=self.__SetDataType__ (Data,DataType)
                            
        if ByteOrder=="":
            self.Images[Index].ByteOrder=self.SysByteOrder
        else:
            self.Images[Index].ByteOrder=ByteOrder
                
        self.Images[Index].StaticHeader["Size"]  = "%d" % self.Images[Index].Size
        self.Images[Index].StaticHeader["Image"] = Index+1
        self.Images[Index].StaticHeader["HeaderID"] = "EH:%06d:000000:000000" % self.Images[Index].StaticHeader["Image"]
        self.Images[Index].StaticHeader["ByteOrder"]=self.Images[Index].ByteOrder
        self.Images[Index].StaticHeader["DataType"]=self.Images[Index].DataType

        
        self.Images[Index].Header={}            
        self.File.seek(0,2)
        StrHeader = "{\n"
        for i in STATIC_HEADER_ELEMENTS:
            if i in list(self.Images[Index].StaticHeader.keys()):
                StrHeader = StrHeader + ("%s = %s ;\n" % (i , self.Images[Index].StaticHeader[i]))
        for i in list(Header.keys()):
            StrHeader = StrHeader + ("%s = %s ;\n" % (i,Header[i]))
            self.Images[Index].Header[i]=Header[i]
        newsize=(((len(StrHeader)+1)/HEADER_BLOCK_SIZE)+1)*HEADER_BLOCK_SIZE -2                   
        StrHeader = str.ljust(StrHeader,newsize)
        StrHeader = StrHeader+"}\n"

        self.Images[Index].HeaderPosition=self.File.tell()    
        self.File.write(StrHeader)
        self.Images[Index].DataPosition=self.File.tell()

        #if self.Images[Index].StaticHeader["ByteOrder"] != self.SysByteOrder:
        if str.upper(self.Images[Index].ByteOrder) != str.upper(self.SysByteOrder):
            self.File.write((Data.byteswap()).tostring())  
        else:
            self.File.write(Data.tostring())
        
        

    ############################################################################
    #Internal Methods
        
    def __GetDefaultNumpyType__(self, EdfType, index=None):
        """ Internal method: returns NumPy type according to Edf type
        """
        return self.GetDefaultNumpyType(EdfType, index)

    def __GetDefaultEdfType__(self, NumpyType):
        """ Internal method: returns Edf type according Numpy type
        """
        if   NumpyType in ["b",   numpy.int8]:            return "SignedByte"
        elif NumpyType in ["B",  numpy.uint8]:            return "UnsignedByte"
        elif NumpyType in ["h",  numpy.int16]:            return "SignedShort"          
        elif NumpyType in ["H", numpy.uint16]:            return "UnsignedShort"          
        elif NumpyType in ["i",  numpy.int32]:            return "SignedInteger"  
        elif NumpyType in ["I", numpy.uint32]:            return "UnsignedInteger"  
        elif NumpyType == "l":
            if sys.platform == 'linux2':
                return "Signed64"           
            else:
                return "SignedLong"           
        elif NumpyType == "L":
            if sys.platform == 'linux2':
                return "Unsigned64"           
            else:
                return "UnsignedLong"
        elif NumpyType == numpy.int64:                     return "Signed64"
        elif NumpyType == numpy.uint64:                    return "Unsigned64"
        elif NumpyType in ["f", numpy.float32]:            return "FloatValue"         
        elif NumpyType in ["d", numpy.float64]:            return "DoubleValue"
        else: raise TypeError("unknown NumpyType %s" % NumpyType)


    def __GetSizeNumpyType__(self, NumpyType):
        """ Internal method: returns size of NumPy's Array Types
        """
        if   NumpyType in ["b",   numpy.int8]:  return 1
        elif NumpyType in ["B",  numpy.uint8]:  return 1
        elif NumpyType in ["h",  numpy.int16]:  return 2          
        elif NumpyType in ["H", numpy.uint16]:  return 2          
        elif NumpyType in ["i",  numpy.int32]:  return 4  
        elif NumpyType in ["I", numpy.uint32]:  return 4  
        elif NumpyType == "l":            
            if sys.platform == 'linux2':
                return 8    #64 bit
            else:
                return 4    #32 bit
        elif NumpyType == "L":
            if sys.platform == 'linux2':
                return 8    #64 bit
            else:
                return 4    #32 bit
        elif NumpyType in ["f", numpy.float32]: return 4         
        elif NumpyType in ["d", numpy.float64]: return 8
        elif NumpyType == "Q":            return 8 #unsigned 64 in 32 bit
        elif NumpyType == "q":            return 8 #signed 64 in 32 bit
        elif NumpyType == numpy.uint64:   return 8
        elif NumpyType == numpy.int64:    return 8
        else: raise TypeError("unknown NumpyType %s" % NumpyType)


    def __SetDataType__ (self, Array, DataType):
        """ Internal method: array type convertion
        """
        # AVOID problems not using FromEdfType= Array.dtype.char
        FromEdfType= Array.dtype
        ToEdfType= self.__GetDefaultNumpyType__(DataType)
        if ToEdfType != FromEdfType:
            aux=Array.astype(self.__GetDefaultNumpyType__(DataType))    
            return aux
        return Array

    def __del__(self):
        try:
            if self.__ownedOpen:
                self.File.close()
        except:
            pass

    def GetDefaultNumpyType(self, EdfType, index=None):
        """ Returns NumPy type according Edf type
        """
        if index is None:return GetDefaultNumpyType(EdfType)
        EdfType=str.upper(EdfType)
        if EdfType in ['SIGNED64']  :return numpy.int64
        if EdfType in ['UNSIGNED64']:return numpy.uint64
        if EdfType in ["SIGNEDLONG", "UNSIGNEDLONG"]:
            dim1 = 1
            dim2 = 1
            dim3 = 1 
            if hasattr(self.Images[index],"Dim1"):
                dim1 = self.Images[index].Dim1
                if hasattr(self.Images[index],"Dim2"):
                    dim2 = self.Images[index].Dim2
                    if dim2 <= 0: dim2 = 1
                    if hasattr(self.Images[index],"Dim3"):
                        dim3 = self.Images[index].Dim3
                        if dim3 <= 0: dim3 = 1
                if hasattr(self.Images[index],"Size"):
                    size = self.Images[index].Size
                    if size/(dim1*dim2*dim3) == 8:
                        if EdfType == "UNSIGNEDLONG":
                            return numpy.uint64
                        else:
                            return numpy.int64
            if EdfType == "UNSIGNEDLONG":
                return numpy.uint32
            else:
                return numpy.int32
            
        return GetDefaultNumpyType(EdfType)
        
    def __readOneHeader(self) :
        buffer = self.File.read(512).decode('utf-8')
        # TEST if it's an EDF File
        if not self.__checkEdfFile.search(buffer) :
            raise IOError("EdfFile : not an Edf format")
        startFind = 0
        while 1:
            endHeader = -1
            while 1:
                endTmp = buffer.rfind('}',startFind,endHeader)
                if endTmp > -1:
                    endHeader = endTmp
                else: break
            if endHeader > -1 : break
            buffer += self.File.read(512).decode('utf-8')
            startFind += 512
        lastnewline = buffer.find('\n',endHeader)
        if endHeader < lastnewline <= endHeader + 2:
            endHeader = lastnewline
        return (buffer,endHeader + 1)

def GetDefaultNumpyType(EdfType):
    """ Returns NumPy type according Edf type
    """
    EdfType=str.upper(EdfType)
    if   EdfType == "SIGNEDBYTE":       return numpy.int8   # "b"
    elif EdfType == "UNSIGNEDBYTE":     return numpy.uint8  # "B"       
    elif EdfType == "SIGNEDSHORT":      return numpy.int16  # "h"
    elif EdfType == "UNSIGNEDSHORT":    return numpy.uint16 # "H"
    elif EdfType == "SIGNEDINTEGER":    return numpy.int32  # "i"
    elif EdfType == "UNSIGNEDINTEGER":  return numpy.uint32 # "I"
    elif EdfType == "SIGNEDLONG":       return numpy.int32  # "i" #ESRF acquisition is made in 32bit
    elif EdfType == "UNSIGNEDLONG":     return numpy.uint32 # "I" #ESRF acquisition is made in 32bit
    elif EdfType == "SIGNED64":         return numpy.int64  # "l"
    elif EdfType == "UNSIGNED64":       return numpy.uint64 # "L"
    elif EdfType == "FLOATVALUE":       return numpy.float32 # "f"
    elif EdfType == "FLOAT":            return numpy.float32 # "f"
    elif EdfType == "DOUBLEVALUE":      return numpy.float64 # "d"
    else: raise TypeError("unknown EdfType %s" % EdfType)


def SetDictCase(Dict, Case, Flag):
    """ Returns dictionary with keys and/or values converted into upper or lowercase
        Dict:   input dictionary
        Case:   LOWER_CASE, UPPER_CASE
        Flag:   KEYS, VALUES or KEYS | VALUES        
    """
    newdict={}
    for i in list(Dict.keys()):
        newkey=i
        newvalue=Dict[i]
        if Flag & KEYS:
            if Case == LOWER_CASE:  newkey = str.lower(newkey)
            else:                   newkey = str.upper(newkey)
        if Flag & VALUES:
            if Case == LOWER_CASE:  newvalue = str.lower(newvalue)
            else:                   newvalue = str.upper(newvalue)
        newdict[newkey]=newvalue
    return newdict    


def GetRegion(Arr,Pos,Size):
    """Returns array with refion of Arr.
       Arr must be 1d, 2d or 3d
       Pos and Size are tuples in the format (x) or (x,y) or (x,y,z)
       Both parameters must have the same size as the dimention of Arr
    """
    Dim=len(Arr.shape)
    if len(Pos) != Dim:  return None
    if len(Size) != Dim: return None
    
    if (Dim==1):
        SizeX=Size[0]
        if SizeX==0: SizeX=Arr.shape[0]-Pos[0]
        ArrRet=numpy.take(Arr, list(range(Pos[0],Pos[0]+SizeX)))
    elif (Dim==2):
        SizeX=Size[0]
        SizeY=Size[1]
        if SizeX==0: SizeX=Arr.shape[1]-Pos[0]
        if SizeY==0: SizeY=Arr.shape[0]-Pos[1]
        ArrRet=numpy.take(Arr, list(range(Pos[1],Pos[1]+SizeY)))
        ArrRet=numpy.take(ArrRet, list(range(Pos[0],Pos[0]+SizeX)),1)
    elif (Dim==3):
        SizeX=Size[0]
        SizeY=Size[1]
        SizeZ=Size[2]
        if SizeX==0: SizeX=Arr.shape[2]-Pos[0]
        if SizeY==0: SizeX=Arr.shape[1]-Pos[1]
        if SizeZ==0: SizeZ=Arr.shape[0]-Pos[2]
        ArrRet=numpy.take(Arr, list(range(Pos[2],Pos[2]+SizeZ)))
        ArrRet=numpy.take(ArrRet, list(range(Pos[1],Pos[1]+SizeY)),1)
        ArrRet=numpy.take(ArrRet, list(range(Pos[0],Pos[0]+SizeX)),2)
    else:
        ArrRet=None
    return ArrRet

##@brief Edf class to read Gzip file
class EdfGzipFile(EdfFile) :
    def __init__(self,filename,**keys) :
        self.__gzipFd = EdfGzipFile._gzFile(filename)
        EdfFile.__init__(self,self.__gzipFd,**keys)

    class _gzFile(gzip.GzipFile) :
         def __init__(self,filepath,*args) :
                gzip.GzipFile.__init__(self,filepath,*args)
                self.name = filepath
                
##@brief Edf class to read Bzip2 file
#
try:
    import bz2
    class EdfBzip2File(EdfFile) :
        def __init__(self,filename,**keys) :
            self.__bzipFd = bz2.BZ2File(filename)
            EdfFile.__init__(self,self.__bzipFd,**keys)
except ImportError: pass

#EXEMPLE CODE:        
if __name__ == "__main__":
    if 1:
        import os
        a = numpy.zeros((5, 10))
        for i in range(5):
            for j in range(10):
                 a[i,j] = 10*i + j 
        edf = EdfFile("armando.edf")
        edf.WriteImage({},a)
        del edf #force to close the file
        inp = EdfFile("armando.edf")
        b = inp.GetData(0)
        out = EdfFile("armando2.edf")
        out.WriteImage({}, b)
        del out #force to close the file
        inp2 = EdfFile("armando2.edf")
        c = inp2.GetData(0)
        print("A SHAPE = ", a.shape)
        print("B SHAPE = ", b.shape)
        print("C SHAPE = ", c.shape)
        for i in range(5):
            print("A", a[i,:])
            print("B", b[i,:])
            print("C", c[i,:])
            
        x = numpy.arange(100)
        x.shape = 5, 20
        for item in ["SignedByte", "UnsignedByte",
                     "SignedShort", "UnsignedShort", 
                     "SignedLong",  "UnsignedLong", 
                     "Signed64", "Unsigned64",
                     "FloatValue", "DoubleValue"]:
            fname = item + ".edf"
            if os.path.exists(fname):
                os.remove(fname)
            towrite = EdfFile(fname)
            towrite.WriteImage({}, x, DataType = item, Append=0)
        sys.exit(0)
    
    #Creates object based on file exe.edf
    exe=EdfFile("images/test_image.edf")
    x=EdfFile("images/test_getdata.edf")
    #Gets unsigned short data, storing in an signed long
    arr=exe.GetData(0,Pos=(100,200),Size=(200,400))
    x.WriteImage({},arr,0)

    arr=exe.GetData(0,Pos=(100,200))
    x.WriteImage({},arr)

    arr=exe.GetData(0,Size=(200,400))
    x.WriteImage({},arr)

    arr=exe.GetData(0)
    x.WriteImage({},arr)
    
    sys.exit()
        
    #Creates object based on file exe.edf
    exe=EdfFile("images/.edf")

    #Creates long array , filled with 0xFFFFFFFF(-1)
    la=numpy.zeros((100,100))
    la=la-1
    
    #Creates a short array, filled with 0xFFFF
    sa=numpy.zeros((100,100))
    sa=sa+0xFFFF
    sa=sa.astype("s")

    #Writes long array, initializing file (append=0)
    exe.WriteImage({},la,0,"")
    
    #Appends short array with new header items
    exe.WriteImage({'Name': 'Alexandre', 'Date': '16/07/2001'},sa)    

    #Appends short array, in Edf type unsigned
    exe.WriteImage({},sa,DataType="UnsignedShort")    

    #Appends short array, in Edf type unsigned
    exe.WriteImage({},sa,DataType="UnsignedLong")    

    #Appends long array as a double, considering unsigned
    exe.WriteImage({},la,DataType="DoubleValue",WriteAsUnsigened=1)

    #Gets unsigned short data, storing in an signed long
    ushort=exe.GetData(2,"SignedLong")

    #Makes an operation
    ushort=ushort-0x10

    #Saves Result as signed long
    exe.WriteImage({},ushort)

    #Saves in the original format (unsigned short)
    OldHeader=exe.GetStaticHeader(2)
    exe.WriteImage({},ushort,1,OldHeader["DataType"] )




import h5py
import numpy as np

def nxsinfo(filename):
    f = h5py.File(filename)
    ct = f['/entry/instrument/detector/count_time/'].value/1000.
    nf = f['/entry/instrument/detector/data/'].shape
    nf = nf[0]
    f.close()
    return float(ct), nf


import numpy as np
from .betaratio import betaratio
from ..Xfit.FitPoissonGammaLikelihood import fit_pg_likelihood


def prob2beta( prob, gproi):
    """Use betaratios to calculate the speckle contrast of a given data set of 
    photon probabilities.
    """
    nq, nbins, ltimes = np.shape(prob[1:,1:])
    beta = np.ma.array(np.empty((nq, nbins-1, ltimes), dtype=np.float32))
    k = prob[0,2:,0]

    for j in range(nq):
        kb = prob[j+1,1]
        beta[j,1:] = betaratio(k, kb, prob[j+1,2:], (ltimes, gproi[j]))[0]
        beta[j,0] = kb
    
    return beta

def prob2betasigma( prob, gproi, sigma=3):
    """Use betaratios to calculate the speckle contrast of a given data set of 
    photon probabilities using only kbar values in a sigma-sigma intercal.
    """
    nq, nbins, ltimes = np.shape(prob[1:,1:])
    beta = np.ma.array(np.empty((nq, nbins-1, 1), dtype=np.float32))
    k = prob[0,2:,0]
    for j in range(nq):
        kb = prob[j+1,1]
        mean_kb = np.mean(kb[kb>0])
        std_kb = np.std(kb[kb>0])
        ind = np.where(np.abs(kb-mean_kb) < std_kb*sigma)[0]
        prob_sel = prob[j+1,2:]
        prob_sel = np.mean(prob_sel[:,ind],-1)
        beta[j,1:] = betaratio(k, mean_kb, prob_sel, (ltimes, gproi[j]))[0]
        beta[j,0] = mean_kb
    
    return beta

def average_beta(t, q, contrast, ratio=0):
    """calculate the speckle contrast of a series of speckle patterns in three ways:
    1) calculating the arithmetik mean (v2_av)
    2) calculating the median (v2_md)
    3) calculating the value with the highest counts (v2_mx)
    """
    v2_av = np.zeros((t.size+1, q.size+1, 2), dtype=np.float32)
    v2_av[1:,0,0] = t
    v2_av[0,1:,0] = q
    v2_md = v2_av.copy()
    v2_mx = v2_av.copy()

    for it in range(t.size):
        for iq in range(q.size):
            beta = contrast[0][it][iq,ratio+1]
            v2_av[it+1,iq+1] = np.ma.mean(beta), np.ma.std(beta)
            v2_md[it+1,iq+1] = np.ma.median(beta), np.ma.std(beta)
            h = np.histogram(beta[~beta.mask], bins=200, range=(-2,8), normed=True)
            e = h[1][:-1] + np.diff(h[1][:2])/2
            v2_mx[it+1,iq+1] = e[np.argmax(h[0])], .05

    return v2_av, v2_md, v2_mx

def beta_from_likelihood(t, q, prob, npix, **kwargs):
    """Use likelihood ratio minimization to estimate the contrast.
    """
    v2_ml = np.zeros((t.size+1, q.size+1, 2), dtype=np.float32)
    v2_ml[1:,0,0] = t
    v2_ml[0,1:,0] = q
    for it in range(t.size):
        for iq in range(q.size):
            probi = prob[0][it][iq+1]
            v2_ml[it+1,iq+1] = fit_pg_likelihood(probi, npix[iq], **kwargs)[0]

    return v2_ml
    



import os
import numpy as np
import pickle
import pandas as pd
from datetime import datetime

from .Decorators import Decorators
from .misc.xsave import save_result, make_filename


class Xdb:
    """Data base class for handling analysis results and making them accessible through the
       data interpretation modules.
    """

    def __init__(self, dbfile=None, db=None, **kwargs):
        self.db = db
        self.dbfile = dbfile
        if dbfile is not None:
            self.load_db(dbfile)
        self.savdir = None

    def __str__(self):
        return ('Xana Instance\n' +
                "savdir: {}\n" +
                "sample name: {}\n" +
                "database file: {}\n" +
                'setup file: {}\n').format(self.savdir, self.sample,
                                           self.dbfile, self.setupfile)

    def __repr__(self):
        return self.__str__()

    # Data Base
    def load_db(self, dbfile=None, init=False, **kwargs):
        path, fname = make_filename(self, 'dbfile', dbfile)
        dbfile = path + fname
        print('Try loading database:\n\t{}'.format(dbfile))
        try:
            self.db = pickle.load(open(dbfile, 'rb'))
            self.dbfile = dbfile
            print('Successfully loaded database')
        except OSError:
            print('\t...loading database failed.')
            if init:
                self.dbfile = self.savdir + os.path.split(dbfile)[-1]
                print(
                    'Analysis database not found.\nInitialize database...'.format(dbfile))
                self.init_db(dbfile, **kwargs)

    def init_db(self, dbfile=None, handle_existing='raise'):
        if self.meta is not None:
            meta_names = list(self.meta.columns)
        else:
            meta_names = []
        names = ['use', 'sample', 'analysis', *meta_names,
                 'mod', 'savname', 'savfile', 'setupfile', 'comment']
        tmp_db = pd.DataFrame(columns=names)
        self.db = tmp_db
        path, fname = make_filename(self, 'dbfile', dbfile)
        self.dbfile = path + fname
        self.save_db(handle_existing=handle_existing)

    def add_db_entry(self, series_id, savfile, method):
        self.db.reset_index(inplace=True, drop=True)
        savname = savfile.split('/')[-1]
        # dbn = self.db.shape[0]
        entry = {'use':True,
                 'sample':self.sample,
                 'analysis':method,
                 'mod':datetime.today(),
                 'savname':savname,
                 'savfile':savfile,
                 'setupfile':self.setupfile,
                 'comment':""
        }
        entry.update(dict(zip(self._meta_save.columns,
                              self._meta_save.loc[series_id],)))
        entry = pd.DataFrame(entry, index=[0])
        self.db = pd.concat([self.db, entry], join='outer', ignore_index=True, sort=False, copy=False)
        self.save_db(handle_existing='overwrite')

    @Decorators.input2list
    def discard_entry(self, db_id, save=True):
        discard = []
        for i in db_id:
            datdir = self.db.loc[i, 'datdir']
            series = self.db.loc[i, 'series']
            sample = self.db.loc[i, 'sample']
            subset = self.db.loc[i].get('subset', np.nan)
            cond = (self.db['datdir'].str.match(datdir))     \
                        & (self.db['series'] == series)         \
                        & (self.db['sample'].str.match(sample)) \

            if not np.isnan(subset):
                cond &= self.db['subset'] == subset
            discard.append(self.db[cond].index.values)

        if len(discard) == 0:
            print('No entry discarded.')
        else:
            self.db.loc[np.unique(np.hstack(discard)), 'use'] = False
            if save:
                self.save_db(handle_existing='overwrite')

    def save_db(self, filename=None, handle_existing='raise'):
        savdir, dbfile = make_filename(self, 'dbfile', filename)
        self.dbfile = save_result(
            self.db, 'Analysis', savdir, dbfile, handle_existing=handle_existing)

    def append_db(self, dbfile, check_duplicates=True):
        if isinstance(dbfile, str):
            path, fname = make_filename(self, filename=dbfile)
            dbfile = path + fname
            if os.path.isfile(dbfile):
                db = pickle.load(open(dbfile, 'r+b'))
            else:
                print('File %s does not exist.' % dbfile)
                return None
        elif isinstance(dbfile, pd.DataFrame):
            db = dbfile
        self.db = pd.concat([self.db, db], join='outer', ignore_index=True, sort=False, copy=False)

        if len(self.db[self.db['use'] == False]) and check_duplicates:
            self.discard_entry(
                self.db[self.db['use'] == False].index.values, save=False)

        self.db.drop_duplicates(inplace=True)
        self.db.reset_index(drop=True, inplace=True)

    def get_item(self, item):
        if type(item) == str:
            return pickle.load(open(item, 'rb'))
        else:
            return pickle.load(open(self.db.loc[item]['savfile'], 'rb'))

    def export_txt(self, name, item, key):
        """Export array to txt file.

        name: filename of exported
        item: db entry to export
        key: key of loaded dictionary to save
        """

        data = self.get_item(item)[key]
        np.savetxt(name, data)
        print('Data saved as ', name)

    @Decorators.input2list
    def rm_db_entry(self, db_id, rmfile=False):
        for i in db_id:
            if rmfile:
                try:
                    os.remove(self.db.loc[i]['savfile'])
                except FileNotFoundError:
                    print('File not found.')
            self.db = self.db.drop(i)
        self.save_db()



#! /usr/bin/env python2.7
import numpy as np
import os
from DropletizeYuri.XPCS_mario import dropimgood, dropimgood_nn
import numpy.ma as ma
import EdfFile
import time

def loadedf(filename,imgn=0):
    f = EdfFile.EdfFile(filename)
    return f.GetData(imgn)

def count(a):
    res = [0]*10
    ct = 1
    for i in range(1,len(a)):
        if a[i]>=1:
            if a[i] == a[i-1]:
                ct += 1
            else:
                if ct<=9:
                    res[ct-1]+=1
                ct = 1
    return res

mf = '/gpfs/exfel/data/user/reiserm/GeO2_ID10/'
if not 'dark' in locals():
     darks = np.load(mf + 'darks.npy')
     dark = np.mean(darks,axis=-1)
if not 'mask' in locals():
     mask = ~np.load(mf + 'mask.npy')
if not 'data' in locals():
    if os.path.isfile(mf + 'tmp_dat.npy'):
        data = np.load(mf + 'tmp_dat.npy')
    else:
        print('Loading data...')
        tmp = loadedf('./data/GeO2_att0_3sec/img_0001.edf')
        data = np.zeros((tmp.shape[0],tmp.shape[1],778), dtype=np.float16)
        for i in range(778):
            filename = './data/GeO2_att0_3sec/img_{0:04d}.edf'.format(i+1)
            data[:,:,i] = loadedf(filename)
            if i%10 == 0:
                print(i)
        print('Data Loaded.')
        np.save(mf + 'tmp_dat.npy', data)
dd = np.load(mf +  'DropDat.npy')#'DropletizeYuri/DropData/GeO2_att0_3sec/''009_DropDat_YuriModified.npy' 
coun = np.load(mf + 'coun_nn2.npy')
pix = np.load(mf + 'pix_nn2.npy')

#_______Main_______
# mint = np.zeros(778)
# for i in range(data.shape[-1]):
#     tmp = data[:,:,i]*1.-dark
#     tmp[tmp<100] = 0.
#     mint[i] = np.mean(tmp[~mask].astype(np.float32))

#Sum and dropletize image and get photon events
darkblank = np.zeros_like(dark)
nx = ny = 1024
gp = nx*ny-np.sum(mask)
print('Found {0} bad pixels'.format(np.sum(mask)))
kbm = []
sp = range(1,52,10)
# for ii,i in enumerate(range(10)):
#     for jj,j in enumerate(range(0,data.shape[-1]-i,20)):
#         ind = np.arange(j,j+i+1)
#         if i>0:
#             sim = np.sum(data[:,:,ind], axis=-1)-dark*(i+1)*1.
#         else:
#             sim = np.squeeze(data[:,:,ind])*1.-dark*1.
#         sim = np.ma.array(sim, mask=mask)
#         imd = dropimgood(sim,darkblank,100*(i+1),1500,2155,75000*(i+1),1900,1024,1024)
#         pb = count(imd[1])
#         #pb = [np.sum(sim.astype(np.float32)) ,gp-sum(pb)] + pb
#         pb = [imd[0] ,gp-sum(pb)] + pb
#         print(pb)
#         kbm.append([pb, imd[:2]])
for ii,i in enumerate(range(10)):
    for jj,j in enumerate(range(0,data.shape[-1]-i,20)):
        ind = np.arange(j,j+i+1)
        tkb = int(np.sum(coun[ind]))
        pb = count(np.sort(pix[ind].ravel()))
        pb = [tkb ,gp-sum(pb)] + pb
        print(pb)
        kbm.append([pb])

np.save(mf + 'kbm_nn2_dts.npy', kbm)
# np.save(mf + 'mint.npy', mint)





    



import numpy as np
import lmfit
from lmfit.model import Model
import corner

def get_ml_solution(res, fix):
    # find the maximum likelihood solution
    highest_prob = np.argmax(res.lnprob)
    hp_loc = np.unravel_index(highest_prob, res.lnprob.shape)
    mle_soln = res.chain[hp_loc]
    fit_report = "\nMaximum likelihood Estimation"
    fit_report += '\n-----------------------------'

    i = 0
    for par in res.params:
        if par in fix:
            fit_report += f'{par} fixed to {fix[par]}'
        else:
            # res.params[par].value = mle_soln[i]
            quantiles = np.percentile(res.flatchain[par], [2.28, 15.9, 50, 84.2, 97.7])
            res.params[par].stderr = 0.5 * (quantiles[3] - quantiles[1])
            fit_report += '\n {} = {} +/- {}'.format(par,res.params[par].value,res.params[par].stderr)
            i += 1

    return res, fit_report

def residual(pars, x, func, data=None, eps=None):
    """2D Residual function to minimize
    """
    model = func.eval(pars, x=x)

    if eps is not None:
        resid = np.sqrt((data - model)**2 * eps**2)
    else:
        resid = np.abs(data - model)
    return resid

def lnlike(pars, x, func, data=None, eps=None):
    v = pars.valuesdict()
    model = func.eval(pars, x=x)
    if eps is not None:
        inv_sigma2 = 1.0/(1/eps**2 + model**2*np.exp(2*np.log(v['f'])))
    else:
        inv_sigma2 = 1.0/(model**2*np.exp(2*np.log(v['f'])))

    return -0.5*(np.sum(residual(pars, x, func, data)**2*inv_sigma2 - np.log(inv_sigma2/(2*np.pi))))

def init_pars(model,init, x, y):
    pars = model.make_params()
    #make initial guess for parameters
    for vn in pars:
        if vn not in init:
            if model.name == 'Model(linear)':
                if vn == 'm':
                    init[vn] = (np.nanmean(np.diff(y)/np.diff(x)), None, None)
                elif vn == 'b':
                    init[vn] = (y[0], None, None)
            if model.name == 'Model(power)':
                if vn == 'a':
                    init[vn] = (y[0], None, None)
                elif vn == 'n':
                    init[vn] = (1, None, None)
                elif vn == 'b':
                    init[vn] = (y[0], None, None)
            if model.name == 'Model(quadratic)':
                if vn == 'a':
                    init[vn] = (np.nanmean(np.diff(y)/(np.diff(x)**2)), None, None)
                elif vn == 'b':
                    init[vn] = (np.nanmean(y), None, None)
                elif vn == 'c':
                    init[vn] = (y[0], None, None)
            if model.name == 'Model(exponential)':
                if vn == 'a':
                    init[vn] = (y[0], None, None)
                elif vn == 't':
                    init[vn] = (x[len(x)//2], 0, None)
                elif vn == 'b':
                    init[vn] = (0, None, None)
                elif vn == 'g':
                    init[vn] = (1, None, None)
        p = init[vn]
        pars[vn].set(p[0], min=p[1], max=p[2])
    return pars


#Defined Models: straight line, power law, quadratic, cubic, exponential
def linear(x, m, b):
    return m*x + b

def power(x, a, n, b):
    return a*x**n + b

def quadratic(x, a, b, c):
    return a*x**2 + b*x + c

def exponential(x, a, t, b, g):
    return a*np.exp(-(x/t)**g) + b

# Main Fit Function
def fit_basic( x, y, dy=None, model='line', init={}, fix=None, emcee=False, plot_corner=False):

    if model in 'linear':
        func = linear
    elif model in 'power':
        func = power
    elif model in 'quadratic':
        func = quadratic
    elif model in 'exponential':
        func = exponential
    else:
        raise ValueError('Model {} not defined.'.format(model))

    model = Model(func, nan_policy='omit')

    pars = init_pars(model, init, x, y)

    if fix is not None:
        for vn in fix:
            pars[vn].set(value=fix[vn], vary=0)

    if dy is not None:
        dy = np.abs(dy)
        wgt = np.array([1./dy[i] if dy[i]>0 else 0 for i in range(len(y))])
        is_weighted = True
    else:
        wgt = None
        is_weighted = False

    if emcee:
        mi = lmfit.minimize(residual, pars, args=(x, model), kws={'data':y, 'eps':wgt},
                            method='nelder', nan_policy='omit')
        # mi.params.add('f', value=1, min=0.001, max=2)
        mini = lmfit.Minimizer(residual, mi.params, fcn_args=(x, model), fcn_kws={'data':y, 'eps':wgt})
        out = mini.emcee(burn=300, steps=1000, thin=20, params=mi.params, is_weighted=is_weighted)
        out, fit_report = get_ml_solution(out, fix)
        print(list(out.params.valuesdict().values()))
        if plot_corner:
            corner.corner(out.flatchain, labels=out.var_names, truths=list(out.params.valuesdict().values()))

    else:
        out = lmfit.minimize(residual, pars, args=(x, model), kws={'data':y, 'eps':wgt}, nan_policy='omit')
        fit_report = lmfit.fit_report(out)

    pars_arr =  np.zeros((len(pars),2))
    for i,vn in enumerate(pars):
        pars_arr[i,0] = out.params[vn].value
        pars_arr[i,1] = out.params[vn].stderr if  out.params[vn].stderr else 0

    if not emcee:
        gof = np.array([out.chisqr, out.redchi, out.bic, out.aic])
    else:
        gof = 0
    return pars_arr, gof, out, fit_report, model




#!/usr/bin/env python


import os
import sys
import setuptools
from setuptools import find_packages
from distutils.command.sdist import sdist


def install_package():

    # load the description from the README.md file
    with open("README.md", "r") as fh:
        long_description = fh.read()


    cmdclass = {'sdist': sdist}
    metadata = dict(
        cmdclass = cmdclass,
        name = 'Xana',
        version = '0.0.11',
        packages = setuptools.find_packages(),
        license = 'MIT',
        author = 'Mario Reiser',
        author_email = 'mario.mkel@gmail.com',
        url = 'https://github.com/reiserm/Xana',
        download_url = "https://pypi.python.org/pypi/Xana",
        keywords = ['data analysis', 'XPCS', 'XSVS', 'SAXS',],
        description="Analysis software for XPCS, XSVS and SAXS data.",
        long_description = long_description,
        long_description_content_type="text/markdown",
        python_requires ='>= 3.6',
        install_requires = [
            'numpy',
            'pandas',
            'lmfit',
            'pyfai',
            'cbf',
            'matplotlib',
            'emcee',
            'corner',
            'h5py',
            'seaborn',
            'ipywidgets',
        ],
        classifiers=[
            'Development Status :: 3 - Alpha',
            'Intended Audience :: Science/Research',
            'Topic :: Scientific/Engineering :: Physics',
            'License :: OSI Approved :: MIT License',
            'Programming Language :: Python :: 3',
            'Programming Language :: Python :: 3.6',
            'Programming Language :: Python :: 3.7',
        ],
    )

    args = sys.argv[1:]

    build_ext = False
    if 'build_ext' in args:
        build_ext = True

    if build_ext:

        try:
            from numpy.distutils.core import setup, Extension
            # setup f2py
            ext = [Extension(name='Xana.XpcsAna.fecorrt3m',
                        sources=['Xana/XpcsAna/fecorrt3m.f90'],
                        f2py_options=['--verbose'])]

            metadata['ext_modules'] = ext
        except ModuleNotFoundError:
            print('Failed building fortran extensions.')
            from setuptools import setup

    else:
        from setuptools import setup

    setup(**metadata)


if __name__ == "__main__":
    install_package()



import numpy as np
from matplotlib import pyplot as plt
from ..Xfit.basic import fitline, fitline0, fitconstant
from ..Xfit.MCMC_straight_line import mcmc_sl
from ..Xfit.fit_basic import fit_basic
from ..Xplot.niceplot import niceplot
from matplotlib.offsetbox import AnchoredText
from matplotlib import ticker


def plot_parameters(pars, parameter, R=250e-9, T=22, fit=None, modes=1, ax=None, marker='o',
                    textbox=False, alpha=1, log='', label=None, ci=0, corner_axes=0, mfc=None,
                    format_ticks=True, cmap=None, init={}, fix=None, viscosity=False,
                    fit_report=False, emcee=False, exc=None, excfit=None, excbad=True,
                    weighted=True, xl=None, xlim=[None,None], ylim=[None,None], **kwargs):

    def getD(eta, err=0):
        dD = 0
        D = kb*(T+273.15)/(6*np.pi*R*eta)
        if err:
            dD = D*err/eta
        return D, dD

    def geteta(D, err=0):
        deta = 0
        eta = kb*(T+273.15)/(6*np.pi*R*D*1e-18)
        if err:
            deta = eta*err/D
        return eta, deta

    def blc(q, L, k, lc):
        def A(q):
            return 4*np.pi/lc*q/k*np.sqrt(1-q**2/(4*k**2))
        return 2*(A(q)*L-1+np.exp(-A(q)*L))/(A(q)*L)**2

    def line(x, p):
        return p[0]*x + p[1]

    def power(x, p):
        return p[0]*x**p[1] + p[2]

    if type(modes) == int:
        modes = np.arange(modes-1, modes)
    else:
        modes = np.array(modes)
        modes -= 1

    if parameter in [0, 'G', 'dispersion', 'tau']:
        name = 't'
    elif parameter in [1, 'kww']:
        name = 'g'
    elif parameter in [2, 'f0', 'ergodicity']:
        name = 'b'

    if 'ax' is None:
        fig, ax = plt.subplots(1, 1, figsize=(9, 4))

    kb = 1.381e-23
    m_unit = {
        'G': 'nm{} s-1'.format(alpha),
        'kww': 'nm{}'.format(alpha),
        'f0': 'nm{}'.format(alpha),
        'tau': 'nm{} s'.format(alpha),
    }
    b_unit = {
        'G': 's-1',
        'kww': '',
        'f0': '',
        'tau': 's'
    }
    y_label = {
        'G': r'$\Gamma (s^{-1})$',
        'kww': 'kww',
        'f0': 'ergodicity',
        'tau': r'$\tau\,(s)$'
    }

    if fit == '' or fit is None:
        dofit = False
    else:
        dofit = True

    qv = pars['q']
    qv = qv**alpha

    # values to be excluded
    iip = np.arange(qv.size)
    iif = iip.copy()
    if exc is not None:
        iip = np.delete(iip, exc)
    if excfit is not None:
        iif = np.delete(iif, np.hstack((excfit)))

    if xl is None:
        x = np.linspace(np.min(qv[iif]), np.max(qv[iif]), 100)
    else:
        x = np.linspace(xl[0], xl[1], 100)

    textstr = ""
    markers = ['^', 'v'] if (len(modes)<3) else ['o']
    for ii, i in enumerate(modes):
        if label is None:
            labstr = 'mode {}: {}'.format(i+1, parameter)
        else:
            labstr = label
        textstr += labstr

    # -------plot decay rates--------
        try:
            y = np.asarray(pars['{}{}'.format(name, i)], dtype=np.float32)
            dy = np.asarray(pars['d{}{}'.format(name, i)], dtype=np.float32)
        except KeyError:
            return np.zeros(5)

        y = np.ma.masked_where(~np.isfinite(y), y)
        dy = np.ma.masked_array(dy, mask=y.mask)

        if parameter == 'G':
            y = 1/y
            dy = y**2*dy
        else:
            pass

        nf = np.where(dy.filled(0) <= 0)[0]
        bad_points = nf.size
        if bad_points:
            print('Found {} points with zero error\n'.format(bad_points))
            if excbad:
                iff = np.array([p for p in iif if p not in nf])
                iip = np.array([p for p in iip if p not in nf])
                print('Excluded bad points.')
                if len(iff)==0 or len(iip)==0:
                    return np.zeros(5)

        color = cmap(ci)
        marker = markers[i]
        ax.errorbar(qv[iip], y[iip], dy[iip], fmt=marker,
                    label=labstr, color=color, mfc=mfc)

        if dofit:
            if fit == 'mcmc_line':
                m, b, f_m, m_ls, b_ls = mcmc_sl(
                    qv[iif], y[iif], dy[iif], doplot=corner_axes)
                # ax[0].plot(x2,m_ls*x2+b_ls)
                m, b = [(x[0], np.mean(x[1:])) for x in (m, b)]
            else:
                res = fit_basic(qv[iif], y[iif], dy[iif],
                                fit, dict(init), fix, emcee)
                fitpar = res[0].astype(np.float32)
                yf = res[4].eval(res[2].params, x=x)

            ax.plot(x, yf, color=color, label=None)

        if parameter in ['G', 'tau']:
            if viscosity:
                power = 1 if (parameter=='G') else -1
                textstr += '\neta = {0[0]:.4g} +/- {0[1]:.2g} [cP]'.format(
                    np.array(geteta(*fitpar[0]))*1e3)**power
        elif parameter == 'f0' and dofit and 't' in res[2].params.keys():
            msd = 1/(2*res[2].params['t'].value)
            dmsd = 2*msd**2*res[2].params['t'].stderr
            r_loc = np.sqrt(6*(msd))
            dr_loc = 6/2/r_loc*dmsd
            textstr += 'localization length: {:.2f} +/- {:.2f} nm\n'.format(
                r_loc, dr_loc)

        if fit_report and dofit:
            print('\n' + textstr)
            print('-'*16)
            print(res[3])

    # if format_ticks:
    #     x_labels = ax.get_xticks()
    #     try:
    #         @ticker.FuncFormatter
    #         def major_formatter(x, pos):
    #             return "{:.2f}".format(x)
    #         ax.ticklabel_format(axis='x', useMathText=True,
    #                             style='sci', scilimits=(0, 0))
    #     except:
    #         pass

    # set style
    if alpha == 1:
        x_lab = r'$\mathrm{q} (nm^{-1})$'
    else:
        x_lab = r'$\mathrm{{q}}^{0} (nm^{{-{0}}})$'.format(alpha)
    ax.set_xlabel(x_lab)
    ax.set_ylabel(y_label[parameter])

    if 'x' in log:
        ax.set_xscale('log')
    if 'y' in log:
        ax.set_yscale('log')

    if textbox:
        at = AnchoredText(textstr, loc=2,)
        ax.add_artist(at)

    ax.legend(loc='best')

    # ax.get_yaxis().get_major_formatter().set_useOffset(False)
    # niceplot(ax,)
    ax.set_xlim(xlim)
    ax.set_ylim(ylim)

    if dofit:
        return res
    else:
        return np.zeros(5)




import numpy as np
from numpy import exp, sqrt, pi
import lmfit

def gaussian(x, **p):
    gauss = 0
    for i in range(p['nm']):
        mst = str(i)
        gauss += (p['a'+mst]/(sqrt(2*pi)*p['sig'+mst])) * exp(-(x-p['cen'+mst])**2 /(2*p['sig'+mst]**2))
    gauss +=  + p['bg']
    return gauss

def fitgaussian( x, y, err=None, mode='standard', nmodes=1, start=None, lb=None, ub=None, doplot=0, h_plot=None, ax=None,
                output='pars', xl=None, ylim=None, color=None):
    """ Fit data with Gaussian peak."""
    
    # initialize parameters
    pars = lmfit.Parameters()
    for i in range(nmodes):
        mst = str(i)
        if i > 0:
            start = np.array(start) *1.1
        pars.add('cen'+mst, value=start[0], min=lb[0], max=ub[0], vary=1)
        pars.add('sig'+mst, value=start[1], min=lb[1], max=ub[1], vary=1)
        pars.add('a'+mst, value=start[2], min=lb[2], max=ub[2], vary=1)
    pars.add('bg', value=1., vary=0)
    pars.add('nm', value=nmodes, vary=0)
    
    if 'bg' in mode:
        if len(start) == 3:
            pars['bg'].set(vary=1)
        elif len(start) == 4:
            pars['bg'].set(value=start[3], min=lb[3], max=ub[3], vary=1)
    if err is not None:
        wgt = 1./err**2
    else:
        wgt = np.ones_like(y)
        
    if 'logx' in mode:
        wgt = 1/np.log10(x)

    mod  = lmfit.Model(gaussian)
    out = mod.fit(y, pars, x=x, weights=wgt)
    
    # plot results
    if doplot:
        if xl is None:
            if ax is None:
                ax = plt.gca()
            xl = ax.get_xlim()
            if xl[0] == 0:
                xl = (np.min(x)*0.9,np.max(x)*1.1)
        xf = np.logspace(np.log10(xl[0]),np.log10(xl[1]),100)
        v = out.best_values
        gf = gaussian(xf, **v)
        if 'legend' in doplot:
            labstr = (r'$\mu: {0[0]:.2g},\, \sigma: {0[1]:.2g},\,'+
                     r' \mathrm{{a}}: {0[2]:.2g},\, \mathrm{{bg}}: {0[3]:.2g}$')
            labstr = labstr.format([v[name] for name in out.var_names])
        else:
            labstr = ''
        
        pl = [] 
        pl.append(ax.plot(xf, gf, '-', label=labstr, linewidth=1))
        if 'data' in doplot:
            pl.append(ax.plot(x, y, 'o', markersize=2.5))
            
        if color is None:
            if h_plot is not None:
                color = h_plot.get_color()
            elif 'data' in doplot:
                color = pl[0][0].get_color()
            else:
                color = 'gray'

        for p in pl:
            p[0].set_color(color)
        
        ax.legend()
        if 'report' in doplot:
            print(out.fit_report())
    if output == 'pars':
        pars =  np.zeros((3*nmodes+1,2))
        for i,vn in enumerate(out.var_names):
            pars[i,0] = out.best_values[vn]
            param = list(out.params.values())
            pars[i,1] = 1.*param[i].stderr
        return pars
    elif output == 'fit':
        return out


from os.path import isfile
from .EdfFile3 import EdfFile, EdfGzipFile

####################################
#--- Standard EDF w/r functions ---#
####################################


def loadedf(filename, imgn=0):
    if isfile(filename):
        if filename.endswith('edf'):
            f = EdfFile(filename)
        elif filename.endswith('edf.gz'):
            f = EdfGzipFile(filename)
        return f.GetData(imgn)
    else:
        print("file ", filename, " does not exist!")
        return 0


def saveedf(filename, data, imgn=0):
    try:
        newf = EdfFile(filename)
        newf.WriteImage({}, data, imgn)
        print("file is saved to ", filename)
        return
    except:
        print("file is not saved!")
        return


def headeredf(filename, imgn=0):
    if isfile(filename):
        if filename.endswith('edf'):
            f = EdfFile(filename)
        elif filename.endswith('edf.gz'):
            f = EdfGzipFile(filename)
        return f.GetHeader(imgn)
    else:
        print("file ", filename, " does not exist!")
        return 0



import numpy as np
import numpy as np
from matplotlib import pyplot as plt
from scipy import fftpack as fp

def cross_correlation(q):
    ind = np.argmin(np.abs(q-0.6))
    inten = I2d[:,ind]
    print('Found %d angles' % len(inten))
    phi = chi.copy()

    N = len(inten)
    lags = np.arange(-N+1, N)

    inten = inten - inten.mean()

    acf = np.correlate(inten, inten, mode='full')
    acf /= (N * inten.std()**2)

    l = len(acf)
    x = np.linspace(-4*np.pi,4*np.pi,l)
    print( N, 'Mean Intensity is: ' + str(mean_int_2))
    figure()
    plt.plot(lags,acf,'-')
    plt.grid()

    fourier = fp.fft(acf)
    xf = fp.fftfreq(l,d=1/(2*360))
    xf = fp.fftshift(xf)
    fourier = fp.fftshift(fourier)
    fourier = 1./l * np.absolute(fourier)
    # fourier = fourier[1:]
    # xf = xf[:l//2]

    plt.figure()
    plt.plot(xf,fourier)
    plt.grid()
    # # plt.savefig(savdir + 'phi_corr_long_fourier.png', format='png',dpi=400)


    # plt.show()



import numpy as np
import copy


def get_soq(Isaxs, setup, Vsaxs=None, nbins=1000):


    sI = np.shape(Isaxs)
    if sI == setup.detector.dim:
        ai = setup.ai
        mask = setup.mask
    elif sI == setup.qsec_dim:
        ai = setup.qsec_ai
        mask = setup.qsec_mask
    else:
        raise ValueError(f'Average image of shape {sI} does not match defined Azimuthal Integrators.')

    if mask is None:
        mask = np.ones_like(sI)
        print('No mask defined.')

    if Vsaxs is None:
        q, ii, e = ai.integrate1d(Isaxs, nbins, mask=~(mask.astype(bool)),
                                  unit='q_nm^-1', error_model='poisson')
    else:
        q, ii, e = ai.integrate1d(Isaxs, nbins, mask=~(mask.astype(bool)),
                                  unit='q_nm^-1', variance=Vsaxs)
    return q, ii, e


def pysaxs(data, load=False, calc_soq=True, **kwargs):

    if load:
        Isaxs = obj.get_item(sid)['Isaxs']
    elif isinstance(data, dict):
        sid = data['sid']
        setup = data['setup']
        Isaxs, Vsaxs = data['get_series'](sid, method='average', **kwargs)
        saxsd = {'Isaxs':Isaxs, 'Vsaxs':Vsaxs}
    else:
        raise ValueError('Could not handle input type during SAXS analysis.')

    tmp = get_soq(Isaxs, setup, Vsaxs)
    soq = np.hstack(tmp)
    soq = soq.reshape(-1, 3, order='F')
    saxsd['soq'] = soq

    return saxsd



import h5py
import numpy as np
import os
import re
from os.path import isfile


#######################################
#--- get information on data series ---#
#######################################    

def getfiles(datdir, suffix, numfmt='(_\d)*'):
    check_suffix = re.compile(suffix)
    filelist = [ os.path.abspath(datdir + item) for item in os.listdir(datdir)
                 if os.path.isfile(os.path.join(datdir, item))
                 and bool(check_suffix.search(item))]
    filelist = sorted(filelist,
                      key=lambda x: int(''.join(re.findall( numfmt ,x.split('/')[-1]))))
    return filelist

def files2series(filelist, masters, seriesfmt,):
    series = []
    series_id = np.empty(len(masters), dtype=np.int32)
    find_seriesid = re.compile(seriesfmt)
    for i,m in enumerate(masters):
        idstr = find_seriesid.search(m).group()
        series_id[i] = int(idstr)
        nblocks = len(re.findall('(_\d{4,})', m))
        searchstr = idstr + '.*(_\d{{4,}}){{{}}}'.format(nblocks-1)
        series.append([x for x in filelist if re.search(searchstr, x) is not None])
    return series, series_id

def get_attrs_from_dict(obj, meta, *args, **kwargs):
            
    def init_meta(p):
        for key, value in p.items():
            meta[key] = np.empty(nfiles, dtype=np.dtype(value[1]))
            
    def get_attr(header, p):
        try:
            attr = p[1](header[p[0]])
        except:
            attr = 0
        return attr

    nfiles = len(meta['master'])
    p = obj.attributes
    init_meta(p)
    for i, m in enumerate(meta.copy()['master']):
        filename = obj.datdir + m
        header = obj.get_header(filename)
        for key, value in p.items():
            meta[key][i] = get_attr(header, value)
                
def get_header_h5(*args, **kwargs):
    return 0

def get_attrs_h5(obj, meta, *args, **kwargs):
            
    def init_meta(p):
        for key, value in p.items():
            meta[key] = np.empty(nfiles, dtype=np.dtype(value[1]))
            
    def get_attr(f, p, key):
        if p[0] == obj.h5opt['data']:
            attr = f[p[0]].attrs[key]
        else:
            if p[0]:
                attr = f[p[0]]
                if attr.dtype.kind == 'V':
                    attr = np.mean(attr.value['value']).astype(np.dtype(p[1]))
                elif attr.dtype.kind == 'f':
                    attr = attr[()]#.astype(np.dtype(p[1]))
                elif attr.dtype.kind == 'u':
                    attr = attr[()]#.astype(np.dtype(p[1]))
                elif attr.dtype.kind == 'i':
                    attr = attr[()]
            else:
                attr = f[obj.h5opt['data']].shape[0]
        return attr

    nfiles = len(meta['master'])
    p = obj.attributes
    init_meta(p)
    for i, m in enumerate(meta.copy()['master']):
        filename = obj.datdir + m
        with h5py.File(filename, 'r', driver=obj.h5opt['driver']) as f:
            for key, value in p.items():
                meta[key][i] = get_attr(f, value, key)

                
'''
AGIPD Methods
'''

class metadata:

    def __init__(self, uid, datdir='./'):
        self.datdir = os.path.abspath(datdir)
        

class run(metadata):

    def __init__(self, runid, rundir, attrdict):
        self.id = runid
        self.rundir = rundir
        self.attrdict = attrdict
        self.ntrains = None
        self.train_ids = None

    def get_

class train(run):

    def __init__(self,):
        self.id = None
        self.npulses = None
        self.pulse_ids = None

class pulse

def get_attrs_agipd(obj, meta, depth, uid):

    def init_meta(p):
        for key, value in p.items():
            meta[key] = np.empty(nfiles, dtype=np.dtype(value[1]))
            
    def get_attr(f, p, key):
        if p[0] == obj.h5opt['data']:
            attr = f[p[0]].attrs[key]
        else:
            if p[0]:
                attr = f[p[0]]
                if attr.dtype.kind == 'V':
                    attr = np.mean(attr.value['value']).astype(np.dtype(p[1]))
                elif attr.dtype.kind == 'f':
                    attr = attr[()]#.astype(np.dtype(p[1]))
                elif attr.dtype.kind == 'u':
                    attr = attr[()]#.astype(np.dtype(p[1]))
                elif attr.dtype.kind == 'i':
                    attr = attr[()]
            else:
                attr = f[obj.h5opt['data']].shape[0]
        return attr

    f_files = getfiles(obj.datdir, obj.suffix, obj.fastname)
    s_files = getzafiles(obj.datdir, obj.suffix, obj.slowname)
    print(f_files,s_files)

    p = obj.attributes
    s = obj.slowdata
    init_meta(p)
    for i, m in enumerate(meta.copy()['master']):
        filename = obj.datdir + m
        with h5py.File(filename, 'r', driver=obj.h5opt['driver']) as f:
            for key, value in p.items():
                meta[key][i] = get_attr(f, value, key)

def common_mode_from_hist(im, searchoffset=50):                        
    bins = np.arange(-searchoffset,searchoffset) 
    hist, _ = np.histogram(im,bins)
    CM = np.where(hist==hist.max())[0][0] - searchoffset
    return CM # has to be subtracted from the tail



import numpy as np
from matplotlib import pyplot as plt
from matplotlib.colors import LogNorm
from ipywidgets import interact, fixed

from readDet.openFile import getEDFimagedata
from readDet.processData import getFilenames


def getimg(run, imgn, fnames, dim, ffmask=0, qroi=0, rgn=0):
    img = getEDFimagedata(fnames[run], first=(imgn,0,0), last=(imgn+1,dim[1],dim[0]), output=0)
    img = np.squeeze(img)
    if rgn:
        img = img[rgn[0]:rgn[1],rgn[2]:rgn[3]]
        if type(ffmask) == np.ndarray:
            ffmask = ffmask[rgn[0]:rgn[1],rgn[2]:rgn[3]]
    if type(ffmask) == np.ndarray:
        img = img * ffmask
        gp = np.sum(ffmask>0)
    else:
        gp = dim[0]*dim[1]
    if qroi:
        for i in range(len(qroi)):
            img[qroi[i]] = np.max(img)
    return img, gp

def browse_imagesEDF(nf=0, run=0, fnames=0, dim=0, exp=0, gp=0, im=0, txt=0, fig=0, ffmask=0, qroi=0, rgn=0):
    def view_image(i=0, **kargs):
        img = getimg(run, i, fnames, dim, ffmask, qroi, rgn)[0]
        im.set_data(img)
        textstr = get_text(img, run, i, exp, gp)
        txt.set_text(textstr)
        fig.canvas.draw_idle()
    interact(view_image, i=(0, nf-1), run=fixed(run), fnames=fixed(fnames), dim=fixed(dim), exp=fixed(exp),
                 gp=fixed(gp), im=fixed(im), txt=fixed(txt), fig=fixed(fig), ffmask=fixed(ffmask), qroi=fixed(qroi),
                 rgn=fixed(rgn))
    
def get_text(img, run, imgn, exp, gp):
    img_sum = np.sum(img)
    img_mean = img_sum / gp
    mx = np.max(img)
    if type(exp) == np.ndarray:
        exp = exp[run]
        ill = imgn*(exp+0.001)
    textstr = (r'$\mathrm{{I}}_{{img}} = {0:.3g}$'+'\n'+ 
           r'$\langle \mathrm{{I}}\rangle _{{pix}} = {1:.3g}$'+'\n'+
           r'$\mathrm{{I}}_{{max}} = {6:.3g}$'+'\n'+
           r'$\mathrm{{t}}_{{exp}} = {2:.3g}s$'+'\n'+ 
           r'$\mathrm{{t}}_{{ill}} = {3:.3g}s$'+'\n'+ 
           r'$\mathrm{{run}}={4:d}$'+'\n'+ 
           r'$\mathrm{{imgn}}={5:d}$').format(img_sum, img_mean, exp, ill, run, imgn, mx)
    return textstr

def plot1imgEDF(run, datdir, imgn, ffmask=0, rgn=0, exp=0, qroi=0, cb=0, norm=None, save=None):
    fnames = datdir
    nf = len(getFilenames(fnames[run],'edf'))
    if type(ffmask) == np.ndarray:
        m,n = np.shape(ffmask)
    else:
        m,n = np.shape(getEDFimagedata(fnames[run], last=(1,)))[1:]
    img, gp = getimg(run, imgn, fnames, (m, n), ffmask, qroi, rgn)
    out = img.copy()
    fig, ax = plt.subplots(figsize=(9,6))
    ax.set_xticklabels('')
    ax.set_yticklabels('')
    ax.tick_params('both', length=0, width=0, which='major')
    if cb:
        im = ax.imshow(img, vmin=cb[0], vmax=cb[1], cmap=plt.get_cmap('magma'), norm=norm)
    else:
        im = ax.imshow(img, cmap=plt.get_cmap('magma'), norm=norm)
    plt.colorbar(im)      
    textstr = get_text(img, run, imgn, exp, gp)
    props = dict(boxstyle='round', facecolor='white')
    txt = ax.text(-0.4, 0.98, textstr, transform=ax.transAxes, fontsize=14,
            verticalalignment='top',horizontalalignment='left', bbox=props)
    propdic = {'nf':nf, 'run':run, 'fnames':fnames, 'dim':(m, n), 'exp':exp, 'gp':gp, 'im':im, 'txt':txt, 'fig':fig,
             'ffmask':ffmask, 'qroi':qroi, 'rgn':rgn}
    if save is None:
        plt.show(block=0)
    else:
        plt.savefig(save, dpi=300)
        plt.close()
    return out, propdic



from os.path import isfile
import cbf

####################################
#--- Standard cbf w/r functions ---#
####################################

def loadcbf(filename):
    if isfile(filename):
        f = cbf.read(filename)
        return f.data
    else:
        print("file ", filename, " does not exist!")
        return 0

def savecbf(filename, data):
    if not isfile(filename):
        cbf.write(filename, data)
    else:
        print("file ", filename, " does already exist!")
    return None

def headercbf(filename):
    if isfile(filename):
        f = cbf.read(filename)
        return f.metadata
    else:
        print("file ", filename, " does not exist!")
        return 0



# Miscellaneous functions used in different modules

def attrstate(d, attr):
    
    if d:
        if attr in d:
            if d[attr]:
                return True
    else:
        return False    

